Require Import CFLib.

(* An assertion [P] is duplicable if [P] implies [P * P * ...]. *)

(* This is a weak definition. A stronger definition would be [P ==> P \* P].
   And another, yet stronger definition would be [P = P \* P]. The latter
   one is too strong; it does not satisfy the lemma [Hexists_duplicable]
   below. The intermediate one is too strong, too; although it works fine
   for many purposes, it does not satisfy the lemma [ghostref_duplicable].
   (It would require a strong premise [op a a = a] instead of a weak one
   [op a a <= a], and that would be too strong when the monoid is [hprop]
   itself.) *)

Definition duplicable (P : hprop) :=
  P ==> P \* P \* (Hexists H, H).

Definition duplicable_post {A : Type} (Q : A -> hprop) :=
  forall x, duplicable (Q x).

Lemma empty_duplicable:
  duplicable \[].
Proof.
  unfold duplicable. xsimpl.
Qed.

Lemma pure_duplicable:
  forall P,
  duplicable \[ P ].
Proof.
  unfold duplicable. intros. xsimpl~.
Qed.

Lemma Hexists_duplicable:
  forall A (Q : A -> hprop),
  (forall x, duplicable (Q x)) ->
  duplicable (Hexists x, Q x).
Proof.
  unfold duplicable. introv H.
  xpull; intros x. xchange (H x). xsimpl.
Qed.

Lemma star_duplicable:
  forall P Q,
  duplicable P ->
  duplicable Q ->
  duplicable (P \* Q).
Proof.
  unfold duplicable. introv HP HQ.
  xchange HP; xpull; intros HG1.
  xchange HQ; xpull; intros HG2.
  xsimpl.
Qed.

Hint Resolve
  empty_duplicable
  pure_duplicable
  Hexists_duplicable
  star_duplicable
: duplicable.

(* The tactic [xgarbage] removes the garbage [Hexists H, H] which (we assume)
   appears in the precondition of the goal. *)

Ltac xgarbage :=
  let garbage := fresh in
  (* This works only if there are no other things to be extracted from the
     precondition. *)
  xpull; intros garbage; xgc garbage.

(* The tactic [xdup H] duplicates the assertion [H], which presumably appears
   in the precondition of the goal. *)

Ltac xdup H :=
  let fact := fresh in
  assert (fact: duplicable H); [
    eauto with duplicable
  | xchange fact; clear fact; try xgarbage ].
