Set Implicit Arguments.
Require Import LibTactics.
Require Import LibStruct.
Require Import LibOrder.

(* -------------------------------------------------------------------------- *)

(* A monoid element [b] is cancellable if [a1 . b = a2 . b] implies [a1 = a2]
   provided the product [a1 . b] is defined. *)

Definition cancellable A (m : monoid_def A) (defined : A -> Prop) (b : A) :=
  let op := monoid_oper m in
  forall a1 a2,
  op a1 b = op a2 b ->
  defined (op a1 b) ->
  a1 = a2.

(* A monoid [m] is cancellative if every element is cancellable. *)

Definition cancellative A (m : monoid_def A) (defined : A -> Prop) :=
  forall b, cancellable m defined b.

(* -------------------------------------------------------------------------- *)

(* A monoid element [a] is idempotent if [a . a] is [a]. *)

Definition idempotent A (m : monoid_def A) (a : A) :=
  let op := monoid_oper m in
  op a a = a.

(* -------------------------------------------------------------------------- *)

(* The ordering that arises out of a monoid is defined by [a <= c] if and only
   if there exists [b] such that [a . b = c]. *)

Section MonoidOrder.

  Context A {m : monoid_def A}.

  Notation op := (monoid_oper m).
  Notation epsilon := (monoid_neutral m).

  Definition monoid_le a c :=
    exists b, op a b = c.

  Global Instance Le_monoid_le : Le A := {
    le := monoid_le
  }.

  Context {M : Monoid m} {MC : Monoid_commutative m}.

  Global Instance Le_preorder_monoid_le : Le_preorder.
  Proof.
    constructor. constructor.
    (* Reflexivity. *)
    { intros a. exists (monoid_neutral m). eapply monoid_neutral_r. }
    (* Transitivity. *)
    { intros b a c [ a' ? ] [ b' ? ]. subst.
      exists (op a' b').
      eapply monoid_assoc. }
  Qed.

  Lemma monoid_le_def_left:
    forall a b,
    a <= op a b.
  Proof.
    intros. exists b. reflexivity.
  Qed.

  Lemma monoid_le_def_right:
    forall a b,
    a <= op b a.
  Proof.
    intros. exists b. eapply monoid_comm.
  Qed.

  Lemma monoid_le_epsilon:
    forall a,
    epsilon <= a.
  Proof.
    intros. exists a. eapply monoid_neutral_l.
  Qed.

  Lemma op_covariant:
    forall a1 b1 a2 b2,
    a1 <= a2 ->
    b1 <= b2 ->
    op a1 b1 <= op a2 b2.
  Proof.
    introv [ a' ? ] [ b' ? ]. subst a2. subst b2.
    exists (op a' b').
    repeat rewrite monoid_assoc. f_equal.
    repeat rewrite <- monoid_assoc. f_equal.
    eapply monoid_comm.
  Qed.

  (* A monoid element [a] is quasi-idempotent if [a . a] is less than [a].
     Note that [a <= a . a] is true by definition of the pre-order [<=],
     so if this pre-order is antisymmetric, then quasi-idempotence and
     idempotence coincide. *)

  Definition quasi_idempotent (a : A) :=
    let op := monoid_oper m in
    op a a <= a.

  Lemma idempotent_quasi_idempotent:
    forall a,
    idempotent m a ->
    quasi_idempotent a.
  Proof.
    unfold idempotent, quasi_idempotent. introv h.
    rewrite <- h at 3. eapply le_refl.
  Qed.

  (* To prove antisymmetry, we seem to need 1- cancellativity and 2- the
     neutral element [epsilon] is atomic, i.e., if it can be written as
     a product, then both members of the product are epsilon itself. *)

End MonoidOrder.

Arguments quasi_idempotent {A} m a.

Hint Resolve idempotent_quasi_idempotent : idempotent.

(* Outside the above section, the notation [a <= b] is accepted by Coq
   only if an appropriate monoid instance is in scope. Coq will not
   automatially apply [Le_monoid_le] to [m], even though both are in
   scope. TEMPORARY? *)

Goal
  forall A (m : monoid_def A) {M : Monoid m},
  let _ := @Le_monoid_le _ m in
  forall a : A,
  a <= a.
Proof.
  intros. eauto using le_refl.
Qed.

(* -------------------------------------------------------------------------- *)

(* A monoid may be equipped with a predicate [defined], which selects a subset
   of ``well-defined'' elements. In the Iris paper, a monoid has a single
   [bottom] element, and an element is defined if and only if it is non-bottom.
   Relying on [defined] instead of [bottom] is roughly equivalent, but can be
   more convenient. *)

Class Monoid_defined A (m : monoid_def A) (defined : A -> Prop) := {

  (* The neutral element [epsilon] must be defined. (This is equivalent,
     in the presence of the next requirement, to the existence of one
     defined element.) *)
  defined_neutral:
    let epsilon := monoid_neutral m in
    defined epsilon;

  (* If a product is defined, then its components must be defined. *)
  defined_split_left:
    let op := monoid_oper m in
    forall a b,
    defined (op a b) ->
    defined a;
}.

(* As a reformulation of [defined_split_left], we find that if [b] is
   defined, then every element [a] smaller than [b] is defined too. *)

Lemma defined_contravariant:
  forall A (m : monoid_def A) (defined : A -> Prop)
  {M : Monoid m} {MD : Monoid_defined m defined},
  let _ := @Le_monoid_le _ m in
  forall a b : A,
  defined b ->
  a <= b ->
  defined a.
Proof.
  introv ? ? ? [ a' ? ]. subst b.
  eauto using defined_split_left.
Qed.

(* Cancellability works for inequations, too. *)

Lemma le_cancel:
  forall A (m : monoid_def A) (defined : A -> Prop)
  {M : Monoid m} {MC : Monoid_commutative m} {MD : Monoid_defined m defined},
  let _ := @Le_monoid_le _ m in
  let op := monoid_oper m in
  forall a1 a2 b,
  op a1 b <= op a2 b ->
  cancellable m defined b ->
  defined (op a2 b) ->
  a1 <= a2.
Proof.
  introv ? ? ? [ f h ] hc hd.
  rewrite <- monoid_assoc in h.
  rewrite (monoid_comm b f) in h.
  exists f.
  eapply hc; rewrite <- monoid_assoc.
  { assumption. }
  { rewrite <- h in hd.  assumption. }
Qed.

(* -------------------------------------------------------------------------- *)

(* Following the Iris paper, once we have a notion of definedness, we define
   compatibility, and from there, (frame-preserving) updateability. *)

Section MonoidUpdate.

  Context A (m : monoid_def A) (defined : A -> Prop) {MD : Monoid_defined m defined}.

  Notation op := (monoid_oper m).

  (* Two elements are compatible iff their product is defined. *)

  Definition compatible a b :=
    defined (op a b).

  (* An element of [a] can be updated to some element of [B] if, for every frame
     [c] such that [a] is compatible with [c], there exists an element [b] of
     [B] that is compatible with [c]. An external choice is implicit in this
     definition: we know that an update from [a] to some element of [B] is
     permitted, but we do not choose which element of [B] we evolve to. *)

  Definition updatable_nondeterministic (a : A) (B : A -> Prop) :=
    forall c,
    compatible a c ->
    exists b,
    B b /\
    compatible b c.

  (* In the special case where [B] is a singleton set, no external choice
     takes place. *)

  Definition updatable a b :=
    updatable_nondeterministic a (eq b).

  (* Every update can be made to preserve definedness. That is, if [a] is
     defined, then an update of [a] to some element of [B] can be required
     to pick a defined element [b] of [B]. *)

  (* At the moment, this lemma is not used explicitly anywhere. It would play a
     role in the metatheory of ghost references; it justifies the fact that, at
     every time, every ghost reference holds a defined element. *)

  Lemma defined_updatable_nondeterministic:
    forall a B,
    defined a ->
    updatable_nondeterministic a B ->
    forall c,
    compatible a c ->
    exists b,
    B b /\
    compatible a c /\
    defined b.
  Proof.
    unfold updatable_nondeterministic. introv hda hc hac.
    forwards (b&?&?): hc hac.
    exists b. splits; eauto.
    unfold compatible in *.
    eapply defined_split_left. eauto.
  Qed.

End MonoidUpdate.

