Set Implicit Arguments.
Require Import LibTactics LibListZ LibInt.
Require Import CFHeaps CFApp CFPrint CFTactics CFDuplicable.
Require Import LibSeq.
Require Import Cascade_ml.

(* -------------------------------------------------------------------------- *)

(* The specification of a cascade. *)

Section Cascade.

(* The type of the elements that we wish to enumerate. *)

Variable A : Type.

(* The cascade invariant. *)

(* For now, we restrict our attention to a cascade without mutable internal
   state, which therefore is duplicable. A cascade may still access a piece of
   external mutable state, described by an assertion [I] that is supplied from
   the outside when the cascade is queried. We require [I] to be invariant;
   this means that several cascades on the same state [I] may exist (and be
   used) at the same time. One could also imagine a scenario where [I] is
   parameterized over the sequence [xs] of elements produced so far, but this
   sounds less useful, since in that case only one cascade would be useable. *)

Variable I : hprop.

(* The predicates [permitted seqs] and [complete seqs] describe the sequences
   of elements that may be produced by this cascade. See [LibSeq]. *)

Variable seqs : sequences A.

(* The predicate [c ~> Cascade xs] means that [c] is a valid cascade, which has
   already produced the elements [xs]. *)

(* This predicate is recursive: indeed, when [c] is queried, it returns a pair
   of an element [x] and a new cascade which has already produced [xs & x].
   Furthermore, in general, a cascade could be infinite, so this predicate is
   co-inductive.

   Instead of using Coq's native co-inductive predicates, we use an
   impredicative encoding. We require an internal description of the cascade,
   that is, a predicate [S] parameterized by [xs] and [c], such that [S] holds
   now and [S] is preserved when the cascade is queried. One can think of [S]
   as a simulation, that is, a witness that we are inside the greatest fixed
   point. Because we wish to restrict our attention to duplicable cascades, we
   require [S] to be duplicable. We still allow [S] to inhabit [hprop], as
   opposed to [Prop], because we have thunks in mind.

   We note that, when [Cascade xs c] holds, then [Cascade] itself is a valid
   witness for [S]. (This is proved below.) So, [Cascade] really is the
   greatest predicate that satisfies these conditions. *)

Definition Cascade (xs : list A) (c : func) : hprop :=
  Hexists S : list A -> func -> hprop,
  (* [S] must hold now. *)
  S xs c \*
  (* [S] must be duplicable. *)
  \[ forall xs c, duplicable (S xs c) ] \*
  (* [S] must imply that [xs] is permitted. *)
  \[ forall xs c, S xs c ==> S xs c \* \[ permitted seqs xs ] ] \*
  (* [S] must be preserved by one step of computation. Note the crucial
     universal quantification on [xs] and [c]. The Hoare triple states that
     querying the cascade requires and preserves [I]. Furthermore, if it
     produces [Nil], then [xs] is complete; if it produces [Cons x c], then
     [c] is a valid cascade which has already produced [xs & x]. *)
  \[
    forall xs c,
    app c [tt]
      INV  (S xs c \* I)
      POST (fun o =>
        match o with
        | Nil =>
            \[ complete seqs xs ]
        | Cons x c =>
            S (xs & x) c
        end)
  ].

(* The above definition can be simplified in the case where [S] is pure. *)

Definition PureCascade (xs : list A) (c : func) : Prop :=
  exists (S : list A -> func -> Prop),
  (forall xs c, S xs c -> permitted seqs xs) /\
  S xs c /\
  (
    forall xs c,
    S xs c ->
    app c [tt]
      INV  I
      POST (fun o => \[
        match o with
        | Nil =>
            complete seqs xs
        | Cons x c =>
            S (xs & x) c
        end])
  ).

(* -------------------------------------------------------------------------- *)

(* This simplified definition helps construct a cascade in the case
   where [S] is pure. See also the tactic [pure_cascade] further on. *)

Theorem pure_cascade:
  forall xs c,
  PureCascade xs c ->
  \[] ==> (c ~> Cascade xs).
Proof using.
  introv (S&Hpermitted&HS&Hspec).
  xunfold Cascade. xsimpl (fun xs c => \[ S xs c ]).
  { clear xs c Hpermitted HS.
    intros xs c. xpull; intros. xapply* Hspec.
    intros [ | x c' ]; xsimpl*. }
  { clear xs c Hspec HS.
    intros xs c. xsimpl*. }
  { clear xs c Hpermitted Hspec HS.
    eauto with duplicable. }
  { clear Hpermitted Hspec.
    exact HS. }
Qed.

(* -------------------------------------------------------------------------- *)

(* A cascade is duplicable. *)

Theorem cascade_duplicable:
  forall c xs,
  duplicable (c ~> Cascade xs).
Proof using.
  intros. xunfold Cascade.
  (* The proof is a little bit tricky because the proof that [S] is duplicable
     is a component of the assertion we are looking at. *)
  eapply Hexists_duplicable; intros S.
  unfold duplicable. xpull; intros.
  xdup (S xs c). { unfold duplicable. eauto. }
  xsimpl*.
Qed.

(* -------------------------------------------------------------------------- *)

(* If [c ~> Cascade xs] holds, then the sequence [xs] is permitted. *)

Theorem cascade_permitted:
  forall c xs,
  c ~> Cascade xs ==>
  c ~> Cascade xs \* \[ permitted seqs xs ].
Proof using.
  intros. xunfold Cascade. xpull; intros S ? Hpermitted ?. xchange* Hpermitted.
Qed.

(* -------------------------------------------------------------------------- *)

(* If [c ~> Cascade xs] holds, then the cascade [c] can be queried
   as follows. *)

Theorem cascade_query:
  forall c xs,
  app c [tt]
    INV  (c ~> Cascade xs \* I)
    POST (fun o =>
      match o with
      | Nil =>
          \[ complete seqs xs ]
      | Cons x c =>
          c ~> Cascade (xs & x)
      end).
Proof using.
  intros.
  xunfold Cascade at 1. unfold Cascade at 1.
  xpull; intros S Hdup Hpermitted Hspec.
  xapply Hspec.
  { xsimpl. xsimpl. }
  { intros [ | x c' ]; xsimpl*; xunfold Cascade; xsimpl*. }
Qed.

(* -------------------------------------------------------------------------- *)

End Cascade.

Ltac pure_cascade :=
  match goal with |- context[?c ~> Cascade ?I ?seqs ?xs] =>
    xchange* (@pure_cascade _ I seqs xs c)
  end.

Hint Resolve cascade_duplicable : duplicable.

(* TEMPORARY add a tactic [xcascade] to help query a cascade. *)
