Set Implicit Arguments.
Require Import LibTactics.
Require Import LibLogic.
Require Import LibSet.
Require Import LibListZ.

(* -------------------------------------------------------------------------- *)

(* A tactic which may move elsewhere. TEMPORARY *)

Lemma split_rev:
  forall A B : Prop,
  B -> A -> A /\ B.
Proof. tauto. Qed.

Ltac break1 :=
  (* Decompose the hypotheses: [unpack] and [branches], combined. *)
  repeat match goal with
  | H: _ /\ _ |- _ => destruct H
  | H: exists a, _ |- _ => destruct H
  | H: _ \/ _ |- _ => destruct H
  end.

Ltac break2 :=
  (* Decompose the goal and try to use our hypotheses to force correct
     instantiation of the meta-variables. If we hit a conjunction, we
     try attacking the subgoals left-to-right and right-to-left, as we
     do not know ahead of time which will lead to a better instantiation
     of the meta-variables. If we hit a choice, we try each side in turn.
     We must require success, otherwise we would make an arbitrary choice. *)
  match goal with
    | |- exists x, _ => eexists; break2
    | |- _ /\ _ => try solve [ split; break2 | eapply split_rev; break2 ]
    | |- _ \/ _ => try solve [ left; break2 | right; break2 ]
    | _ => rew_list; eauto
    end.

Ltac break :=
  break1; subst; break2.

(* -------------------------------------------------------------------------- *)

(* Specifying the permitted behaviors of a producer. *)

(* Suppose we have a (possibly non-deterministic) producer of a sequence of
   elements. Suppose we have a consumer, who is a (finite) observer of this
   sequence. When queried, the producer emits either one more element or an
   end-of-sequence condition. *)

(* We wish to specify which observations are permitted. Such a specification
   must indicate which sequences of elements are valid partial observations
   (i.e., these sequences can be produced) and which sequences of elements are
   valid complete observations (i.e., these sequences, followed with an
   end-of-sequence signal, can be produced). *)

(* In Coq, such a specification could take one of several forms. For instance,
   one could define a type of observations (or: behaviors) and let a
   specification be a predicate on observations. This approach seems
   interesting insofar as it might naturally deal with more complex
   observations (e.g., an observation of divergence or failure). For the
   moment, we follow Filliâtre and Pereira and let a specification be a pair
   of two predicates on sequences. The predicate [permitted] describes which
   partial observations are valid, whereas the predicate [complete] describes
   which complete observations are valid. Thus, [permitted xs] means that the
   producer may produce the elements [xs]. (It may stop there, or may produce
   more.) [complete xs] means that the producer may produce the elements [xs]
   and stop. (This does not imply that it must stop after producing [xs]. The
   fact that a producer must stop after producing [xs] is encoded by the
   property that no sequence of the form [xs & x] is permitted.) *)

(* Ideally, specifications should be viewed as an abstract type. How they are
   implemented in Coq should not affect how they are used. *)

Section Spec.

Variable A : Type.

(* -------------------------------------------------------------------------- *)

(* TEMPORARY move to LibList *)

Implicit Types xs ys : list A.

Definition prefix_closed (P : list A -> Prop) :=
  forall xs ys,
  P ys ->
  prefix xs ys ->
  P xs.

Lemma prefix_closed_nil:
  forall (P : list A -> Prop) xs,
  prefix_closed P ->
  P xs ->
  P nil.
Proof using.
  eauto with prefix.
Qed.

Definition prefix_closure (P : list A -> Prop) : list A -> Prop :=
  fun xs => exists ys, prefix xs ys /\ P ys.

Definition prefix_closure_alt (P : list A -> Prop) : list A -> Prop :=
  fun xs => exists ys, P (xs ++ ys).

Lemma prefix_closure_eq:
  forall P,
  prefix_closure P = prefix_closure_alt P.
Proof using.
  intros. extens; intros xs; split; unfold prefix_closure, prefix_closure_alt, prefix.
  { introv (ys&(zs&?)&?). subst. eauto. }
  { introv (ys&?). eauto. }
Qed.

Lemma prefix_closure_is_prefix_closed:
  forall P,
  prefix_closed (prefix_closure P).
Proof.
  unfold prefix_closed, prefix_closure.
  introv (zs&?&?). eauto using prefix_transitive.
Qed.

(* The relation [fun xs => prefix xs ys] is the prefix closure of the
   singleton [ys]. Thus, it is prefix-closed. *)

Lemma prefix_closure_singleton:
  forall ys : list A,
  (fun xs => prefix xs ys) = prefix_closure (= ys).
Proof using.
  intros. extens; intros xs. unfold prefix_closure.
  split. eauto. intros (?&?&?). subst. eauto.
Qed.

Lemma prefix_closed_prefix:
  forall ys : list A,
  prefix_closed (fun xs => prefix xs ys).
Proof using.
  intros. rewrite prefix_closure_singleton.
  eapply prefix_closure_is_prefix_closed.
Qed.

(* -------------------------------------------------------------------------- *)

(* A specification describes a set of permitted behaviors of the producer. It
   is implemented as a pair of two predicates [permitted] and [complete]. *)

Record sequences :=
  Seqs {
    permitted: list A -> Prop;
    complete: list A -> Prop
  }.

(* The following well-formedness conditions should in principle hold, but are
   not imposed a priori (that would be inconvenient). *)

Definition wf seqs :=
  (* [permitted] is nonempty. *)
  permitted seqs nil /\
  (* [permitted] is prefix-closed. *)
  prefix_closed (permitted seqs) /\
  (* [complete xs] implies [permitted xs]. *)
  forall xs, complete seqs xs -> permitted seqs xs.

Lemma wf_corollary:
  forall seqs,
  wf seqs ->
  forall xs,
  prefix_closure (complete seqs) xs ->
  permitted seqs xs.
Proof using.
  unfold wf, prefix_closure. intros. break.
Qed.

(* -------------------------------------------------------------------------- *)

(* A specification is may-finite if and only if every permitted sequence is a
   prefix of some complete sequence. Intuitively, this means that there is
   always hope that the sequence will end. The predicate [permitted] is then
   exactly the prefix closure of the predicate [complete]. *)

Definition may_finite seqs :=
  forall xs,
  permitted seqs xs ->
  prefix_closure (complete seqs) xs.

Lemma may_finite_corollary:
  forall seqs,
  wf seqs ->
  may_finite seqs ->
  permitted seqs = prefix_closure (complete seqs).
Proof using.
  intros. extens; intros xs. split; eauto using wf_corollary.
Qed.

(* -------------------------------------------------------------------------- *)

(* A deterministic specification allows one list, [ys], and just one list. *)

Definition deterministic (ys : list A) :=
  Seqs
    (fun xs => prefix xs ys)
    (fun xs => xs = ys).

Lemma wf_deterministic:
  forall ys,
  wf (deterministic ys).
Proof using.
  intros ys. splits; simpl; intros; subst;
  eauto using prefix_closed_prefix with prefix.
Qed.

Lemma may_finite_deterministic:
  forall ys,
  may_finite (deterministic ys).
Proof using.
  intros ys xs. simpl. unfold prefix_closure. eauto.
Qed.

(* -------------------------------------------------------------------------- *)

(* The specification [nil] allows the empty list, and just the empty list. *)

Definition epsilon :=
  deterministic nil.

(* -------------------------------------------------------------------------- *)

(* The specification [cons x seqs] allows [x] followed by whatever the
   specification [seqs] allows. *)

Definition cons x seqs :=
  Seqs
    (fun xs => xs = nil \/ exists xs', xs = x :: xs' /\ permitted seqs xs')
    (fun xs => exists xs', xs = x :: xs' /\ complete seqs xs').

Lemma wf_cons:
  forall x seqs,
  wf seqs ->
  wf (cons x seqs).
Proof using.
  introv (Hnil & Hprefix & Hincl). splits; simpl.
  (* Nonemptyness. *)
  { eauto. }
  (* Prefix-closedness. *)
  { intros xs zs [ ? | (ys&?&?) ] ?; subst zs.
    { eauto using prefix_nil_inverse. }
    { forwards [ ? | ? ]: prefix_cons_inverse; unpack; eauto. }}
  (* Inclusion. *)
  { introv (xs'&?&?). eauto. }
Qed.

Lemma deterministic_cons:
  forall x xs,
  cons x (deterministic xs) = deterministic (x :: xs).
Proof using.
  intros. unfold cons. simpl. unfold deterministic. f_equal;
  extens; intros ys; split; intros; try branches; unpack; subst;
  eauto using prefix_cons_inverse with prefix.
Qed.

(* -------------------------------------------------------------------------- *)

(* The specification [singleton x] allows the singleton sequence [xs]. *)

Definition singleton x :=
  deterministic (x :: nil).

Lemma singleton_is_cons_epsilon:
  forall x,
  singleton x = cons x epsilon.
Proof using.
  intros. unfold epsilon. rewrite deterministic_cons. eauto.
Qed.

(* -------------------------------------------------------------------------- *)

(* Concatenation of specifications. *)

Definition concat seqs1 seqs2 :=
  Seqs
    (fun xs =>
      permitted seqs1 xs \/
      exists xs1 xs2, xs = xs1 ++ xs2 /\ complete seqs1 xs1 /\ permitted seqs2 xs2)
    (fun xs =>
      exists xs1 xs2, xs = xs1 ++ xs2 /\ complete seqs1 xs1 /\ complete seqs2 xs2).

Lemma wf_concat:
  forall seqs1 seqs2,
  wf seqs1 ->
  wf seqs2 ->
  wf (concat seqs1 seqs2).
Proof using.
  introv (Hnil1 & Hprefix1 & Hincl1) (Hnil2 & Hprefix2 & Hincl2). splits; simpl.
  (* Nonemptyness. *)
  { eauto. }
  (* Prefix-closedness. *)
  { intros xs ys [ ? | (xs1&xs2&?&?&?) ] Hprefix.
    { eauto. }
    { subst ys.
      forwards [ ? | (xs1a&?&?) ]: prefix_app_r_inverse. { eauto. }
      { eauto. }
      { subst xs. eauto 8. }
    }
  }
  (* Inclusion. *)
  { intros xs (xs1&xs2&?&?&?). subst xs. right. eauto 8. }
Qed.

Lemma concat_left_unit:
  forall seqs,
  wf seqs ->
  concat epsilon seqs = seqs.
Proof using.
  intros (permitted&complete) (Hnil&Hprefix&Hincl). unfold concat, epsilon. simpl in *.
  f_equal; extens; intros xs; split; intros; try branches.
  { forwards: prefix_nil_inverse. eauto. subst.
    eauto with prefix. }
  { unpack. subst. rew_list. eauto. }
  { right. exists (@nil A) xs. rew_list. eauto. }
  { unpack. subst. rew_list. eauto. }
  { exists (@nil A) xs. rew_list. eauto. }
Qed.

Lemma concat_right_unit:
  forall seqs,
  wf seqs ->
  concat seqs epsilon = seqs.
Proof using.
  intros (permitted&complete) (Hprefix&Hincl). unfold concat, epsilon. simpl in *.
  f_equal; extens; intros xs; split; intros; try branches.
  { eauto. }
  { unpack. subst.
    forwards: prefix_nil_inverse. eauto. subst. rew_list.
    eauto. }
  { eauto. }
  { unpack. subst. rew_list. eauto. }
  { exists xs (@nil A). rew_list. eauto. }
Qed.

Lemma concat_associative:
  forall seqs1 seqs2 seqs3,
  concat (concat seqs1 seqs2) seqs3 = concat seqs1 (concat seqs2 seqs3).
Proof using.
  intros (permitted1&complete1) (permitted2&complete2) (permitted3&complete3).
  unfold concat. simpl. f_equal; extens; intros; split; intros; break.
Qed.

Lemma concat_singleton_is_cons:
  forall x seqs,
  wf seqs ->
  concat (singleton x) seqs = cons x seqs.
Proof using.
  intros x (permitted&complete) (Hnil&?&?).
  unfold singleton, deterministic, concat, cons. simpl in *.
  f_equal; extens; intros; split; intros; break.
  { forwards: prefix_cons_inverse. eauto. break. }
  { eauto with prefix. }
Qed.

Lemma deterministic_concat:
  forall xs1 xs2,
  concat (deterministic xs1) (deterministic xs2) = deterministic (xs1 ++ xs2).
Proof using.
  intros. unfold concat, deterministic. simpl.
  f_equal; extens; intros; split; intros; break; eauto with prefix.
  forwards: prefix_app_r_inverse; eauto. break.
Qed.

Lemma fold_cons:
  forall xs seqs,
  wf seqs ->
  List.fold_right cons seqs xs = concat (deterministic xs) seqs.
Proof using.
  induction xs; intros; simpl.
  { rewrite concat_left_unit by eauto. eauto. }
  { rewrite IHxs by eauto.
    rewrite <- concat_singleton_is_cons by eauto using wf_concat, wf_deterministic.
    rewrite <- concat_associative.
    rewrite concat_singleton_is_cons by eauto using wf_deterministic.
    rewrite deterministic_cons.
    eauto. }
Qed.

(* -------------------------------------------------------------------------- *)

(* [elements E] allows the elements of the set [E] to be produced in an
   arbitrary order. This is a nondeterministic specification. *)

Definition elements E :=
  Seqs
    (* A partial list [xs] is permitted if it corresponds to a subset [E']
       of [E]. This implies that [xs] has no duplicate elements. *)
    (fun xs => exists E', list_repr E' xs /\ E' \c E)
    (* A list [xs] is complete if it corresponds to the set [E]. *)
    (fun xs => list_repr E xs).

Lemma wf_elements:
  forall E,
  wf (elements E).
Proof using.
  intro. unfold wf. splits; simpl.
  { eauto using empty_incl, list_repr_nil. }
  { intros xs ys (E'&?&?) ?. exists (to_set xs). split.
    { eauto using list_repr_to_set, no_duplicates_prefix, list_repr_no_duplicates. }
    { eapply incl_trans; eauto.
      forwards: list_repr_to_set_inverse; eauto. subst.
      eauto using prefix_to_set. }
  }
  { intros. exists E. split. eauto. eapply LibBag.incl_refl. }
Qed.

(* -------------------------------------------------------------------------- *)

End Spec.

(* TODO:

   define may-infinite (existence of an infinite increasing chain of permitted sequences)
   define must-finite as the negation of may-infinite?
   define must-infinite?

   if [xs] is permitted and not complete, then there should exist [x]
   such that [xs & x] is permitted; make that part of [wf]?

   define a notion of refinement (inclusion) between specifications?
     show that iterating on a list refines iterating on a set
     show that [iter_specification] is covariant, i.e., if an [iter]
       function implements a precise spec, then it also implements a
       less deterministic spec.

*)
