Set Implicit Arguments.
Require Import Coq.Setoids.Setoid.    (* [rewrite] *)
Require Import Coq.Classes.Morphisms. (* [Proper] *)
(* Load TLC. *)
Require Import LibTactics.
Require Import LibPowerOfTwo.
Require Import LibEquiv.
Require Import HashTable_model.
(* Load the CFML library. *)
Require Import CFLib.
Require Import CFDuplicable.
Require Import CFIteration.
Require Import LibSeq.
Require Import Cascade_spec.
Require Import LibListZ.
(* Load the CFML specs of the OCaml modules [Pervasives] and [Array]. *)
Require Import Pervasives_proof.
Require Import Array_proof.
(* Load the CFML characteristic formulae generated for the OCaml module
   [HashTable]. *)
Require Import HashTable_ml.

(* TEMPORARY move to CFML *)

Definition computes A B (f : func) (F : A -> B) :=
  curried 1 f /\
  forall x,
  app f [x]
    PRE  \[]
    POST \[= F x ].

Definition decides A1 A2 (f : func) (R : A1 -> A2 -> Prop) :=
  curried 2 f /\
  forall a1 a2,
  app f [a1 a2]
    PRE  \[]
    POST \[= isTrue (R a1 a2)].

(* -------------------------------------------------------------------------- *)

(* We define the signature [HashedTypeSpec] as an extension of [HashedType_ml],
   using Coq's [Include] directive. An alternative approach would be to use
   Coq's [Declare Module], making [HashedType_ml] a component of the signature
   [HashedTypeSpec]. The current (flat) approach seems simpler, but the nested
   approach would offer the advantage of allowing finer-grained [Import]s. *)

Module Type HashedTypeSpec.

  Include HashedType_ml.

  Notation key := t_.

  (* There must exist an equivalence relation [E]... *)

  Parameter E : key -> key -> Prop.

  Parameter Eequiv : equiv E.

  (* ...and a mathematical function [H]... *)

  Parameter H : key -> int.

  (* ...such that equivalent keys have the same hash... *)

  Parameter compatibility :
    Proper (E ==> eq) H.

  (* ...and [equal] decides equivalence... *)

  Parameter equal_spec:
    decides equal E.

  (* ...and [hash] computes a hash. *)

  Parameter hash_spec:
    computes hash H.

  Hint Extern 1 (RegisterSpec equal) => Provide equal_spec.
  Hint Extern 1 (RegisterSpec hash) => Provide hash_spec.

End HashedTypeSpec.

(* -------------------------------------------------------------------------- *)

(* The header of the functor [HashTable.Make]. *)

Module MakeSpec (K : HashedTypeSpec).

Import K.

Module MK := Make_ml(K).

(* -------------------------------------------------------------------------- *)

(* [E] is an equivalence relation, and [hash] is compatible with it. *)

Local Hint Resolve Eequiv (Erefl Eequiv) (Esym Eequiv) (Etrans Eequiv) : equiv.

Existing Instance compatibility.

Notation is_domain := (is_domain E).
Notation population := (population E).
Notation removal := (removal E).

(* -------------------------------------------------------------------------- *)

(* Set up local hints. *)

Local Hint Resolve power_of_2_positive power_of_2_double.
Local Hint Resolve land_in_bounds_left land_in_bounds_right.
Local Hint Resolve index_length_unfold int_index_prove. (* for array indexing *)
Local Hint Resolve length_nonneg.

(* Arithmetic trivialities. *)

Lemma foo: forall a b, a < b -> a < 0 + b.  Proof. intros. math. Qed.
Lemma bar: forall n, 0 < n -> 0 <= n.       Proof. intros. math. Qed.
Lemma baz: forall n, 0 <= n -> n >= 0.      Proof. intros. math. Qed.
Lemma qux: forall a b, a <= b - 1 -> a < b. Proof. intros. math. Qed.
Lemma fiz: forall a b, a < b -> a < (b - 1) + 1. Proof. intros. math. Qed.
Lemma hah: forall a, 0 <= a -> 0 <= a *2.   Proof. math. Qed.

Local Hint Resolve foo bar baz qux fiz hah.

(* -------------------------------------------------------------------------- *)

(* CFML convenience. *) (* TEMPORARY *)

(* TEMPORARY *)
Ltac xxunfold foo := xunfold foo at 1; unfold foo at 1; xpull.

(* -------------------------------------------------------------------------- *)

(* Abstract everything over the type [A]. *)

Section S.

Variable A : Type.

Notation empty := (@HashTable_model.empty key A).
Notation add := (HashTable_model.add E).
Notation remove := (HashTable_model.remove E).

(* Buckets are inhabited. This is necessary for TLC's map-access op. a[i]. *)
Local Instance : Inhab (MK.bucket_ A).
Proof using. econstructor. exists (@MK.Void A). eauto. Qed.

(* -------------------------------------------------------------------------- *)

(* Logical functions on buckets. (Internal.) *)

(* A bucket is isomorphic to a list of pairs. One might wish to exhibit this
   isomorphism and exploit existing functions and lemmas about lists. Instead,
   we prefer to develop a little machinery to reason directly about buckets. *)

Implicit Type b : MK.bucket_ A.

(* Buckets are acyclic. A bucket cannot be its own tail. *)

Lemma buckets_are_acyclic:
  forall b k x,
  b = MK.More k x b ->
  False.
Proof using.
  induction b; intros.
  congruence.
  match goal with h: _ = _ |- _ => injection h; clear h; intros h _ _ end. eauto.
Qed.

Hint Resolve buckets_are_acyclic : equiv. (* TEMPORARY hack for decide_if *)

(* [bfilter k b] keeps from [b] only the key-value pairs whose key is equal to
   [k], up to the equivalence relation [E]. It produces a list of values. *)

Fixpoint bfilter k b : list A :=
  match b with
  | MK.Void =>
      nil
  | MK.More k' x b =>
      If E k' k then x :: bfilter k b else bfilter k b
  end.

(* [bremove k b] removes from [b] the first key-value pair whose key is equal to
   [k], up to the equivalence relation [E]. It produces a new bucket. *)

Fixpoint bremove k b : MK.bucket_ A :=
  match b with
  | MK.Void =>
      MK.Void
  | MK.More k' x b =>
      If E k' k then b else MK.More k' x (bremove k b)
  end.

(* Lemmas about [bfilter] and [bremove]. *)

Lemma tail_bfilter:
  forall k' k b,
  E k' k ->
  tail (bfilter k' b) = bfilter k' (bremove k b).
Proof.
  induction b; simpl; intros; eauto. case_if.
  { decide_if. eauto. }
  { decide_if. simpl. decide_if. eauto. }
Qed.

Lemma bfilter_bremove:
  forall k' k b,
  ~ E k' k ->
  bfilter k' b = bfilter k' (bremove k b).
Proof.
  induction b; simpl; intros; eauto. case_if.
  { decide_if. simpl. decide_if. f_equal. eauto. }
  { case_if. eauto. simpl. decide_if. eauto. }
Qed.

Lemma bfilter_bremove_2 k b : bremove k b = b -> bfilter k b = nil.
Proof. induction b; simpl; eauto. intros. decide_if. eapply IHb. congruence. Qed.

Lemma bfilter_bremove_3 k b : bfilter k b = nil -> bremove k b = b.
Proof. induction b; simpl; eauto. intros. decide_if. rewrite IHb; eauto. Qed.

Lemma bfilter_bremove_eq k b : (bremove k b = b) = (bfilter k b = nil).
Proof. extens. iff; eauto using bfilter_bremove_2, bfilter_bremove_3. Qed.

Instance : Proper (E ++> eq ++> eq) bfilter.
Proof.
  intros k k' ?. intros b b' ?; subst b'.
  induction b; simpl; intros; eauto.
  case_if; decide_if; try rewrite IHb; eauto.
Qed.

(* Although (as explained above) we have attempted to avoid exploiting the
   isomorphism between buckets and lists, this isomorphism turns out to be
   useful in the specification of [fold_aux]. One could avoid it, but that
   would be rather artificial. So, here is the injective mapping of buckets to
   lists. *)

Fixpoint bucket2list b : list (key * A) :=
  match b with
  | MK.Void =>
      nil
  | MK.More k x b =>
      (k, x) :: bucket2list b
  end.

(* TEMPORARY keep or throw away?
Lemma bucket2list_injective:
  forall b1 b2, bucket2list b1 = bucket2list b2 -> b1 = b2.
Proof using.
  induction b1; destruct b2; simpl; intros h; try congruence.
  injection h. intros. subst. f_equal. eauto.
Qed.

Lemma bucket2list_injective_eq:
  forall b1 b2, (b1 = b2) = (bucket2list b1 = bucket2list b2).
Proof using.
  extens. split; intros. congruence. eauto using bucket2list_injective.
Qed.
*)

Lemma bucket2list_injective_nil:
  forall b, bucket2list b = nil -> b = MK.Void.
Proof using.
  destruct b; simpl; intros; congruence.
Qed.

(* This ad hoc lemma states that if the bucket [b] is nonempty -- i.e.,
   its list view begins with a key pair [k, x] and continues with [xs] --
   then this bucket must be of the form [MK.More k x b'], where [b'] is
   in fact [bremove k b]. Therefore [xs] is in fact [bucket2list b'],
   and [bfilter k b] is a nonempty list whose head is [x]. *)

Lemma nonempty_bucket:
  forall k x b xs,
  bucket2list b = (k, x) :: xs ->
  b = MK.More k x (bremove k b) /\
  bucket2list (bremove k b) = xs /\
  Some x = head (bfilter k b).
Proof.
  introv h. destruct b; simpl in *.
  { congruence. }
  { injection h. clear h. intros. subst. decide_if. eauto. }
Qed.

(* -------------------------------------------------------------------------- *)

(* The invariant. *)

Implicit Type h : MK.table_ A.
  (* a synonym for [loc] -- but mentioning [A] here is important *)
Implicit Type M : model key A.
Implicit Type data : list (MK.bucket_ A).
Implicit Type d : loc.

Notation "k '//' data" := (Z.land (H k) (length data - 1))
  (at level 40).

(* The table entries that concern the key [k] should be stored in the bucket
   at index [k // data], and nowhere else. This is expressed by the predicates
   [content M data] and [no_garbage data]. Together, these predicates
   characterize [bfilter k data[i]] for every index [i]. *)

Definition content M data :=
  forall k,
  bfilter k data[k // data] = M k.

Definition no_garbage data :=
  forall k i,
  0 <= i < length data ->
  i <> k // data ->
  bfilter k data[i] = nil.

(* [TableRecord] is just a notation for the ownership of the record. *)

Notation TableRecord d population init :=
  (`{
    MK.data' := (d : loc);
    MK.population' := (population : int);
    MK.init' := (init : int)
  }).

(* The predicate [table_inv M init data] expresses every pure invariant,
   except the constraint on the [population] field. It is convenient to leave
   that constraint out, as it is sometimes temporarily violated. *)

Definition table_inv M init data :=
  power_of_2 (length data) /\
  power_of_2 init /\
  content M data /\
  no_garbage data /\
  (exists D, is_domain D M)
.

Hint Unfold table_inv : table.

(* The (public, abstract) predicate [h ~> TableInState M s] means that there
   is a valid hash table at address [h], whose logical model is [M], and whose
   concrete state is represented by [s]. The type of [s] is [state], which in
   the client's eyes is an abstract type. This gives us the ability of giving
   an abstract name to the current state of a table. Therefore, this allows us
   to express the fact that certain operations do not modify the table (for
   instance, reading, iterating, etc.), while others do. This in turn allows
   us to express the fact that certain operations require an unmodified table
   (e.g., producing the next element of an iterator). *)

Definition state : Type :=
  loc * list (MK.bucket_ A).
  (* Internally, a state [s] is a pair [d, data] which records the address and
     content of the data array. *)

Implicit Type s : state.

Definition TableInState M s h :=
  Hexists d pop init data,
  h ~> TableRecord d pop init \*
  d ~> Array data \*
  \[ table_inv M init data ] \*
  \[ population M = pop ] \*
  \[ s = (d, data) ].

(* The (public, abstract) predicate [h ~> Table M] means that there
   is a valid hash table at address [h], whose logical model is [M]. *)

Definition Table M h :=
  Hexists s, h ~> TableInState M s.

(* -------------------------------------------------------------------------- *)

(* Exposing/wrapping up variables / hypotheses. *)

(* Opening [Table], revealing just [TableInState]. *)
Ltac intros1 := let s := fresh "s" in intros s.
Ltac expose1 := xxunfold Table; intros1.
(* Opening [TableInState], revealing just [table_inv]. *)
Ltac intros2 := intros d pop init data Htable Hpop Hs;
                try match goal with h: ?s = (?d, ?data) |- _ => clear s h end.
Ltac expose2  := xxunfold TableInState; intros2.
(* Opening [table_inv]. *)
Ltac expose3 := match goal with h: table_inv _ _ _ |- _ =>
                destruct h as (Hdata & Hinit & Hcontent & Hnogarbage & (D&Hdomain)) end.
(* Opening [Table] (if present) and [TableInState] and [table_inv]. *)
Ltac expose  := first [ expose1; expose2 | expose2 ]; expose3.

(* Opening a predicate that appears in the postcondition.
   This is used when concluding. *)
Ltac wrapup3 := unfold table_inv.
Ltac wrapup2 := xunfold TableInState.
Ltac wrapup1 := xunfold Table.
Ltac wrapup  := wrapup1; wrapup2; wrapup3.

(* -------------------------------------------------------------------------- *)

(* It is convenient to be able to read certain fields of the data structure
   without having to unfold the predicate [TableInState]. We prove a few ad
   hoc lemmas for this purpose. *)

(* TEMPORARY use everywhere (implicitly if possible) *)

(* Reading the length of the array. *)

Lemma in_state_length_spec:
  forall M h s d data,
  s = (d, data) ->
  app Array_ml.length [d]
    INV  (h ~> TableInState M s)
    POST \[= length data ].
Proof using.
  intros. rename d into d'. rename data into data'. expose2.
  assert (d' = d). { congruence. } subst d'.
  assert (data' = data). { congruence. } subst data'.
  xstraight. wrapup. xsimpl*.
Qed.

(* Reading a slot of the array. *)

Lemma in_state_bucket_read_spec:
  forall M h s d data i,
  s = (d, data) ->
  index data i ->
  app Array_ml.get [d i]
    INV  (h ~> TableInState M s)
    POST \[= data[i] ].
Proof using.
  intros. rename d into d'. rename data into data'. expose2.
  assert (d' = d). { congruence. } subst d'.
  assert (data' = data). { congruence. } subst data'.
  xstraight. wrapup. xsimpl*.
Qed.

(* -------------------------------------------------------------------------- *)

(* A table can be converted to a read-only view (for a fresh state [s]), and
   back. *)

Lemma name_current_state:
  forall h M,
  h ~> Table M ==>
  Hexists s, h ~> TableInState M s.
Proof using.
  intros. expose1. xsimpl*.
Qed.

Lemma forget_current_state:
  forall h M s,
  h ~> TableInState M s ==>
  h ~> Table M.
Proof using.
  intros. wrapup1. xsimpl*.
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [content]. *)

Lemma content_proper:
  forall M data,
  content M data ->
  Proper (E ==> eq) M.
Proof.
  introv Hcontent. intros k k' h.
  do 2 rewrite <- Hcontent.
  rewrite h. reflexivity.
Qed.

Lemma content_empty:
  forall n,
  power_of_2 n ->
  content empty (make n MK.Void).
Proof using.
  intros. intros k.
  rewrite read_make by (rewrite length_make; eauto).
  reflexivity.
Qed.

Lemma content_add:
  forall M k x data,
  power_of_2 (length data) ->
  content M data ->
  content
    (add M k x)
    data[k // data := MK.More k x data[k // data]].
Proof using.
  unfold content, add. introv Hdata Hcontent. intros k'.
  rewrite length_update.
  case_if as HE.
  (* Case: [k'] is logically equal to [k]. They must have the same hash. *)
  { rewrite (compatibility HE).
    rewrite read_update_case by eauto. decide_if.
    simpl. decide_if.
    rewrite <- Hcontent.
    rewrite (compatibility HE).
    reflexivity. }
  (* Case: [k'] and [k] are unrelated. *)
  { rewrite read_update_case by eauto. case_if as Hb.
    (* Sub-case: [k'] and [k] fall into the same bucket, by chance. *)
    { simpl. decide_if. rewrite Hb. eapply Hcontent. }
    (* Sub-case: [k'] and [k] fall into distinct buckets. *)
    { eapply Hcontent. }
  }
Qed.

Lemma content_remove:
  forall M data k,
  power_of_2 (length data) ->
  content M data ->
  content
    (remove M k)
    data[k // data := bremove k data[k // data]].
Proof using.
  unfold content, remove. introv Hdata Hcontent. intros k'.
  rewrite length_update.
  case_if as HE.
  (* Case: [k'] is logically equal to [k]. They must have the same hash. *)
  { rewrite (compatibility HE).
    rewrite read_update_case by eauto. decide_if.
    rewrite <- Hcontent.
    rewrite (compatibility HE).
    erewrite tail_bfilter by eauto.
    reflexivity. }
  (* Case: [k'] and [k] are unrelated. *)
  { rewrite read_update_case by eauto. case_if as Hb.
    (* Sub-case: [k'] and [k] fall into the same bucket, by chance. *)
    { rewrite Hb. rewrite <- Hcontent. erewrite <- bfilter_bremove by eauto. reflexivity. }
    (* Sub-case: [k'] and [k] fall into distinct buckets. *)
    { eapply Hcontent. }
  }
Qed.

Hint Resolve
  content_empty
  content_add
  content_remove
  content_proper
: content.

(* TEMPORARY *)
Lemma tail_self_inv_eq:
  forall (xs : list A),
  (tail xs = xs) = (xs = nil).
Proof.
  intros. extens. iff.
  { eauto using tail_self_inv. }
  { subst. eauto. }
Qed.

Lemma content_remove_unchanged: (* TEMPORARY used? rename? move? *)
  forall M data k,
  power_of_2 (length data) ->
  content M data ->
  (bremove k data[k // data] = data[k // data]) =
  (remove M k k = M k).
Proof.
  unfold content, remove. introv Hdata Hcontent.
  decide_if.
  rewrite tail_self_inv_eq.
  rewrite <- Hcontent.
  eapply bfilter_bremove_eq.
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [no_garbage]. *)

Lemma no_garbage_contrapositive:
  forall data k i x,
  no_garbage data ->
  0 <= i < length data ->
  Some x = head (bfilter k data[i]) ->
  i = k // data.
Proof using.
  introv h1 h2 h3.
  tests ht: (i = k // data). { eauto. } false.
  specializes h1 h2 ht.
  rewrite h1 in h3. false.
Qed.

Lemma no_garbage_empty:
  forall n,
  0 <= n ->
  no_garbage (make n MK.Void).
Proof using.
  intros n ? k i Hi ?.
  rewrite length_make in Hi by eauto.
  rewrite read_make by eauto.
  reflexivity.
Qed.

Lemma no_garbage_add:
  forall k x data,
  no_garbage data ->
  power_of_2 (length data) ->
  no_garbage data[k // data := MK.More k x data[k // data]].
Proof using.
  unfold no_garbage. introv Hnogarbage ?. intros k' i Hi Hdiff.
  rewrite length_update in *.
  rewrite read_update_case by eauto. case_if; [ | eauto ].
  simpl. clear Hi. case_if as HE; [ false | eauto ].
  rewrite HE in Hdiff. eauto.
Qed.

Lemma no_garbage_remove:
  forall data k,
  no_garbage data ->
  no_garbage data[k // data := bremove k data[k // data]].
Proof using.
  unfold no_garbage. introv Hnogarbage. intros k' i Hi Hdiff.
  rewrite length_update in *.
  rewrite read_update_case by eauto. case_if; [ | eauto ].
  assert (~ E k' k). { intro h. forwards: compatibility h. congruence. }
  rewrite <- bfilter_bremove by eauto. eauto.
Qed.

Hint Resolve
  no_garbage_empty
  no_garbage_add
  no_garbage_remove
: no_garbage.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [table_inv]. *)

Lemma table_inv_empty:
  forall init n,
  power_of_2 init ->
  power_of_2 n ->
  table_inv empty init (make n MK.Void).
Proof.
  intros. unfold table_inv.
  splits; try rewrite length_make by eauto;
  eauto with equiv content no_garbage is_domain.
Qed.

Hint Resolve table_inv_empty : table.

(* -------------------------------------------------------------------------- *)

(* Public properties of [Table]. *)

(* If [h ~> Table M] holds, then [M] has finite domain and is compatible
   with [E]. These properties ensure that [population M] is meaningful. *)

Theorem Table_properties:
  forall h M,
  h ~> Table M ==> h ~> Table M \* \[
    (exists D, is_domain D M) /\
    Proper (E ==> eq) M
  ].
Proof using.
  intros. expose. wrapup. xsimpl*; splits; eauto using content_proper.
Qed.

(* -------------------------------------------------------------------------- *)

Lemma le_reflexive: forall x : int, x <= x.
Proof. intros. eapply preorder_refl. eapply le_preorder. Qed.
Hint Resolve le_reflexive.

Lemma power_2_above_spec_preliminary:
  forall (k : nat) (x n : int),
  n <= x * 2 ^ k ->
  power_of_2 x ->
  app MK.power_2_above [x n]
    PRE  \[]
    POST (fun y => \[ power_of_2 y /\ x <= y ]).
Proof using.
  induction k; rew_maths; simpl; intros x n Hnx Hx.
  (* Case: [0]. *)
  { rewrite Z.mul_1_r in Hnx. xcf. xgo; try solve [ eauto | false; math ]. }
  (* Case: [k + 1]. *)
  { rewrite Zpow_facts.Zpower_Zsucc in Hnx by apply Zle_0_nat.
    rewrite Z.mul_assoc in Hnx.
    xcf. xgo; intuition eauto using power_of_2_double_larger. }
Qed.

Lemma power_2_above_spec:
  forall x n : int,
  power_of_2 x ->
  app MK.power_2_above [x n]
    PRE  \[]
    POST (fun y => \[ power_of_2 y /\ x <= y ]).
Proof using.
  intros x n Hx.
  forwards [ zk [ Hzk Hn ]]: (@powers_of_2_diverge x n).
  { eauto. }
  forwards [ k Hk ]: Z_of_nat_complete Hzk. subst zk.
  eauto using power_2_above_spec_preliminary.
Qed.

Hint Extern 1 (RegisterSpec MK.power_2_above) => Provide power_2_above_spec.

(* -------------------------------------------------------------------------- *)

Lemma create_spec:
  forall init : int,
  app MK.create [init]
    PRE  \[]
    POST (fun h => h ~> Table empty).
Proof using.
  assert (power_of_2 16).
  { exists 4. split. math. reflexivity. }
  xcf. xstraight. intros. wrapup1. wrapup2.
  xsimpl; eauto with equiv population table.
Qed.

Hint Extern 1 (RegisterSpec MK.create) => (Provide create_spec).

(* -------------------------------------------------------------------------- *)

(* TEMPORARY *)
Lemma make_ext:
  forall A `{Inhab A} x (xs : list A),
  (forall i, 0 <= i < length xs -> xs[i] = x) ->
  xs = make (length xs) x.
Proof using.
  intros. eapply ext_eq.
  { symmetry. eauto using length_make, length_nonneg. }
  { intros i Hi. rewrite read_make by eauto. eauto. }
Qed.

Lemma clear_spec:
  forall h M,
  app MK.clear [h]
    PRE  (  h ~> Table M)
    POST (#(h ~> Table empty)).
Proof using.
  xcf. expose. xstraight.
  xpull. intros data' Hdata'. rewrite <- Hdata' in *.
  intros Hcontent' _.
  (* The table is now logically empty. *)
  forwards Hempty: make_ext data'. eauto.
  rewrite Hempty.
  wrapup1. wrapup2. xsimpl; eauto with equiv population table.
Qed.

(* -------------------------------------------------------------------------- *)

Lemma reset_spec:
  forall h M,
  app MK.reset [h]
    PRE  (  h ~> Table M)
    POST (#(h ~> Table empty)).
Proof using.
  xcf. expose. xstraight.
  intros _. wrapup1. wrapup2. xsimpl; eauto with equiv population table.
Qed.

(* -------------------------------------------------------------------------- *)

Lemma copy_spec_ro:
  forall h M s,
  app MK.copy [h]
    INV  (h ~> TableInState M s)
    POST (fun h' => h' ~> Table M).
Proof using.
  xcf. expose2. xstraight. intros h'. wrapup. xsimpl*.
Qed.

Lemma copy_spec:
  forall h M,
  app MK.copy [h]
    INV  (h ~> Table M)
    POST (fun h' => h' ~> Table M).
Proof using.
  intros.
  xchange* name_current_state; xpull; intros s.
  xapp_spec copy_spec_ro.
  xchange* (@forget_current_state h M s).
Qed.

(* -------------------------------------------------------------------------- *)

Lemma population_spec_ro:
  forall h M s,
  app MK.population [h]
    INV  (h ~> TableInState M s)
    POST \[= population M ].
Proof using.
  xcf. expose2. xstraight. wrapup. xsimpl*.
Qed.

Lemma population_spec:
  forall M h,
  app MK.population [h]
    INV  (h ~> Table M)
    POST \[= population M ].
Proof using.
  intros.
  xchange* name_current_state; xpull; intros s.
  xapp_spec population_spec_ro.
  xchange* (@forget_current_state h M s).
Qed.

(* [length] is a synonym for [population]. *)

Lemma length_spec:
  MK.length = MK.population.
Proof using.
  intros. xcf_show. eauto.
Qed.

(* -------------------------------------------------------------------------- *)

Lemma index_spec:
  forall h k pop d init data,
  app MK.index [h k]
    INV  (h ~> TableRecord d pop init \*
          d ~> Array data)
    POST (fun i => \[ i = k // data ]).
Proof using.
  xcf. xstraight.
Qed.

Hint Extern 1 (RegisterSpec MK.index) => (Provide index_spec).

(* -------------------------------------------------------------------------- *)

Lemma resize_aux_spec:
  forall M init data,
  table_inv M init data ->
  forall h d pop b,
  app MK.resize_aux [h b]
    PRE  (h ~> TableRecord d pop init \* d ~> Array data)
    POST (# Hexists M' data',
          h ~> TableRecord d pop init \*
          d ~> Array data' \*
          \[ table_inv M' init data' ] \*
          \[ length data' = length data ] \*
          \[ forall k, M' k = bfilter k b ++ M k ]
    ).
Proof using.
  induction b; intros; xcf; xmatch.
  (* Case: [Void]. *)
  { xret. xsimpl* (>> M data). }
  (* Case: [More]. *)
  { (* The recursive call. *)
    xapp*. clear IHb. intros M' data' Htable Hlen Hpost.
    rewrite <- Hlen.
    (* The addition. This is an inlined copy of [add], without the population
       update and without the call to [resize], of course. *)
    expose3. xstraight.
    (* The postcondition. *)
    wrapup3. xsimpl; try splits; repeat rewrite length_update;
    eauto with equiv content no_garbage is_domain.
    { unfold add. intros. rewrite Hpost.
      case_if; decide_if; try rewrite app_cons; reflexivity. }
  }
Qed.

(* -------------------------------------------------------------------------- *)

Lemma resize_spec:
  forall M h,
  app MK.resize [h]
    INV  (  h ~> Table M)
    POST (# \[]).
Proof using.
  xcf. expose1. expose2. xstraight.
  xif; [ | xret; wrapup; xsimpl* ].
  xstraight.
  match goal with foo: array _ |- _ => rename foo into ndata end.
  (* Give the loop invariant. *)
  expose3.
  xfor_inv (fun i =>
    Hexists M' data',
    d ~> Array data \*
    h ~> TableRecord ndata pop init \*
    ndata ~> Array data' \*
    \[ table_inv M' init data' ] \*
    \[ length data * 2 = length data' ] \*
    \[ forall k, M' k = If k // data < i then M k else nil ]
  ).
  (* Indices go up. *)
  { eauto. }
  (* The loop invariant is true initially. *)
  { xsimpl empty; subst; eauto with table.
    { intros k. case_if; [ | reflexivity ].
      false. forwards: (@land_in_bounds_left (H k) _ Hdata). math. }
    { rewrite length_make by eauto. eauto. }
  }
  (* The loop invariant is preserved. *)
  { intros i [ ? ? ]. xpull. intros M0 data0 Htable0 Hnsize0 HM0.
    xstraight.
    xapp_spec resize_aux_spec. { eauto. }
    xpull. intros M1 data1 Htable1 Hnsize1 HM1.
    xsimpl; [ intros k | congruence | eauto ].
    rewrite HM1, HM0.
    do 2 case_if; try solve [ false; math ]. (* TEMPORARY 3 cas, dont deux semblables, a unifier *)
    { forwards Hempty: Hnogarbage k i.
      { eauto. }
      { math. }
      rewrite Hempty. reflexivity. }
    { assert (here: k // data = i). { math. }
      unfold content in Hcontent.
      rewrite app_nil_r, <- Hcontent, here.
      reflexivity. }
    { forwards Hempty: Hnogarbage k i.
      { eauto. }
      { math. }
      rewrite Hempty. reflexivity. }
  }
  (* The loop invariant finally implies the postcondition. *)
  { intros _. xpull. intros M' data1 Htable' Hnsize1 HM'.
    (* We have achieved a copy! *)
    assert (Hcopy: M' = M).
    { unfold model in *. extens. intros k.
      rewrite HM'. case_if; [ reflexivity | ].
      false. forwards: (@land_in_bounds_right (H k) _ Hdata). math. }
    rewrite Hcopy in Htable'. wrapup1. wrapup2. xsimpl*.
  }
Qed.

Hint Extern 1 (RegisterSpec MK.resize) => (Provide resize_spec).

(* -------------------------------------------------------------------------- *)

Lemma increment_population_spec:
  forall M h d pop init data,
  table_inv M init data ->
  population M = pop + 1 ->
  app MK.increment_population [h]
    PRE  (  h ~> TableRecord d pop init \* d ~> Array data)
    POST (# h ~> Table M).
Proof using.
  introv Htable Hpop. xcf. expose3. xstraight.
  (* By incrementing [h.population], we have re-established [Table]. *)
  xpre (h ~> Table M). { wrapup. xsimpl*. }
  xif; xstraight.
Qed.

Hint Extern 1 (RegisterSpec MK.increment_population) => (Provide increment_population_spec).

(* -------------------------------------------------------------------------- *)

Theorem add_spec:
  forall M h k x,
  app MK.add [h k x]
    PRE  (h ~> Table M)
    POST (# Hexists M',
          h ~> Table M' \*
          \[ M' = add M k x ] \*
          \[ lean M -> M k = nil -> lean M' ]). (* redundant, but possibly helpful *)
Proof using.
  xcf. expose. xstraight. xapp*.
  (* The preconditions of [increment_population]. *)
  { wrapup3. splits; try rewrite length_update;
    eauto with equiv content no_garbage is_domain. }
  { erewrite population_add by eauto with equiv content. math. }
  (* The redundant postcondition. *)
  { xsimpl*. eauto using lean_add with equiv content. }
Qed.

(* -------------------------------------------------------------------------- *)

Lemma find_all_aux_spec:
  forall k b,
  app MK.find_all_aux [k b]
    PRE  \[]
    POST \[= bfilter k b].
Proof using.
  induction b; xcf; xmatch; xstraight.
  (* Case: [More]. *)
  xif; xstraight; xsimpl; decide_if; eauto.
Qed.

Hint Extern 1 (RegisterSpec MK.find_all_aux) => (Provide find_all_aux_spec).

Lemma find_all_spec_ro:
  forall M h s k,
  app MK.find_all [h k]
    INV  (h ~> TableInState M s)
    POST (fun xs => \[ xs = M k ]).
Proof using.
  xcf. expose. xstraight. wrapup. xsimpl*.
Qed.

Lemma find_all_spec:
  forall M h k,
  app MK.find_all [h k]
    INV  (h ~> Table M)
    POST (fun xs => \[ xs = M k ]).
Proof using.
  intros.
  xchange* name_current_state; xpull; intros s.
  xapp_spec find_all_spec_ro.
  xchange* (@forget_current_state h M s).
Qed.

(* -------------------------------------------------------------------------- *)

Lemma find_aux_spec:
  forall k b,
  app MK.find_aux [k b]
    PRE  \[]
    POST \[= head (bfilter k b)].
Proof using.
  intros. induction b;
  xcf; xmatch; simpl; xstraight.
  xif; xstraight; decide_if; xsimpl*.
Qed.

Hint Extern 1 (RegisterSpec MK.find_aux) => (Provide find_aux_spec).

Lemma find_spec_ro:
  forall M h s k,
  app MK.find [h k]
    INV  (h ~> TableInState M s)
    POST \[= head (M k) ].
Proof using.
  xcf. expose. xstraight.
  wrapup. xsimpl*. rewrite <- Hcontent. assumption.
Qed.

Hint Extern 1 (RegisterSpec MK.find) => (Provide find_spec_ro).

Lemma find_spec:
  forall M h k,
  app MK.find [h k]
    INV  (h ~> Table M)
    POST \[= head (M k) ].
Proof using.
  intros.
  xchange* name_current_state; xpull; intros s.
  xapp.
  xchange* (@forget_current_state h M s).
Qed.

(* -------------------------------------------------------------------------- *)

(* TEMPORARY move *)
Lemma simplify_cons_cons_equality:
  forall k x b1 b2,
  (MK.More k x b1 = MK.More k x b2) = (b1 = b2).
Proof using.
  intros. extens. split; intros; congruence.
Qed.

Lemma remove_aux_spec:
  forall h k b d pop init,
  app MK.remove_aux [h k b]
    PRE  (h ~> TableRecord d pop init)
    POST (fun b' => Hexists pop',
          h ~> TableRecord d pop' init \*
          \[ b' = bremove k b /\
             pop' = (If b' = b then pop else pop - 1) ]).
Proof using.
  induction b; intros; xcf; xmatch; xstraight.
  (* Case: [Void]. *)
  { xsimpl*. decide_if. eauto. }
  (* Case: [More]. *)
  xif; decide_if; xstraight.
  (* Sub-case: key match. *)
  { xsimpl*; decide_if; eauto. (* buckets_are_acyclic *) }
  (* Sub-case: key mismatch. *)
  xapp_spec infix_eq_eq_gen_spec. introv Heq. (* TEMPORARY [xapp_spec foo as x] is missing? *)
  xif; xret; do 2 rewrite simplify_cons_cons_equality.
  (* Sub-sub-case: physical equality. *)
  { rew_refl in Heq. do 2 decide_if. xsimpl*. }
  (* Sub-sub-case: physical disequality. *)
  { xsimpl*. }
Qed.

Hint Extern 1 (RegisterSpec MK.remove_aux) => (Provide remove_aux_spec).

Lemma remove_spec:
  forall M h k,
  app MK.remove [h k]
    PRE  (h ~> Table M)
    POST (# Hexists M',
          h ~> Table M' \*
          \[ M' = remove M k ] \*
          \[ lean M -> lean M' ]). (* redundant, but possibly helpful *)
Proof using.
  xcf. expose. xstraight.
  erewrite content_remove_unchanged by eauto.
  intros _. wrapup. subst.
  xsimpl* (remove M k).
  { eauto using lean_remove with equiv content. }
  { eauto with equiv content population. }
  { rewrite length_update.
    splits; eauto with equiv content no_garbage population is_domain. }
Qed.

(* -------------------------------------------------------------------------- *)

Lemma replace_spec:
  forall M h k x,
  app MK.replace [h k x]
    PRE  (h ~> Table M)
    POST (# Hexists M',
          h ~> Table M' \*
          \[ M' = add (remove M k) k x ] \*
          \[ lean M -> lean M' ]). (* redundant, but possibly helpful *)
Proof using.
  xcf. expose. xstraight.
  erewrite content_remove_unchanged by eauto.
  (* We have just performed a [remove] followed by an [add], with a few
     write/read and write/write optimizations. Argue. *)
  forwards HcontentR: content_remove k; eauto.
  forwards HcontentA: content_add k x HcontentR; try rewrite length_update; eauto.
  forwards HnogarbageR: no_garbage_remove k; eauto.
  forwards HnogarbageA: no_garbage_add k x; try rewrite length_update; eauto.
  clear HcontentR HnogarbageR.
  rewrite length_update in HcontentA, HnogarbageA.
  rewrite read_update_eq in HcontentA, HnogarbageA by eauto.
  rewrite update_update_eq in HcontentA, HnogarbageA by eauto.
  (* Call [increment_population]. *)
  xapp*.
  { wrapup3. splits; try rewrite length_update;
    eauto with equiv content is_domain. }
  { erewrite population_add by eauto with equiv content is_domain.
    erewrite population_remove by eauto with equiv content is_domain.
    subst. eauto. }
  (* The redundant postcondition. *)
  { xsimpl*. eauto using lean_replace with equiv content. }
Qed.

(* -------------------------------------------------------------------------- *)

Lemma mem_spec_ro:
  forall M h s k,
  app MK.mem [h k]
    INV  (h ~> TableInState M s)
    POST \[= isTrue (M k <> nil) ].
Proof using.
  xcf. xstraight. xmatch; xret; xsimpl; eauto using use_None_head, use_Some_head.
Qed.

Lemma mem_spec:
  forall M h k,
  app MK.mem [h k]
    INV  (h ~> Table M)
    POST \[= isTrue (M k <> nil) ].
Proof using.
  intros.
  xchange* name_current_state; xpull; intros s.
  xapp_spec mem_spec_ro.
  xchange* (@forget_current_state h M s).
Qed.

(* -------------------------------------------------------------------------- *)

(* The auxiliary recursive function [fold_aux f b] enumerates the key-value
   pairs in the bucket [b], in left-to-right order. In other words, it
   enumerates the sequence [bucket2list b]. In order to facilitate the proof
   by induction on [b], we give a direct specification, where the invariant
   [I] is parameterized by the list of remaining elements. *)

Lemma fold_aux_spec_preliminary:
  forall B : Type,
  forall I : list (key * A) -> B -> hprop,
  forall f,
  (
    forall kx kxs accu,
    cca_unboxed f kx accu
      (            I (kx :: kxs) accu)
      (fun accu => I        kxs  accu)
  ) ->
  forall b accu,
  app MK.fold_aux [f b accu]
    (  I (bucket2list b) accu)
    (fun accu => I nil   accu).
Proof using.
  introv Hf. induction b; simpl; xcf; xval; subst; xmatch.
  { xret. xsimpl. }
  { (* TEMPORARY bad choice of names by [xmatch] *)
    match goal with xx : A |- _ => rename xx into x end.
    xlet. eapply (Hf (k, x)). xapp. }
Qed.

(* We can then derive another specification of [fold_aux], which is an
   instance of [Fold]. This time, the invariant [I] is
   parameterized by the elements that have been enumerated already. *)

Lemma fold_aux_spec:
  forall b,
  let permitted := fun kxs => prefix kxs (bucket2list b) in
  let complete  := fun kxs => kxs = bucket2list b in
  forall B (I : list (key * A) -> B -> hprop),
  Fold MK.fold_aux (@cca_unboxed _ _ _)
    permitted complete I
    (fun b' => \[ b' = b ]) (fun b' => \[]).
Proof using.
  unfold Fold. introv Hf. intros accu. xpull. xintros.
  xapp_spec fold_aux_spec_preliminary (fun kys accu =>
    Hexists kxs, \[ kxs ++ kys = bucket2list b ] \* I kxs accu).
  { intros. xpull. intros kys Heq.
    xapply (Hf kx kys).
    { rewrite <- Heq. exists kxs. rew_list. eauto. }
    { xsimpl. }
    { xsimpl. rewrite <- Heq. rew_list. eauto. }
  }
  { rew_list. eauto. }
  { xsimpl*. rew_list in *. subst. eauto with prefix. }
Qed.

(* The specification of [fold] is necessarily somewhat complicated. Indeed, the
   order in which the key-value pairs are presented to the client is partly
   unspecified. It is not completely unspecified, though: if the client has
   added two values [x1] and [x2] for the same key [k], then it is guaranteed
   that [fold] will present the most-recently-added key-value pair first.

   A concise way of explaining to the client in which order the key-value pairs
   are presented is as follows. [fold] behaves AS IF at each iteration a
   key-value pair [k, x] was removed from the table (for an arbitrary key [k])
   and presented to the client -- until the table is empty.

   In other words, the permitted partial sequences [kxs] are those that satisfy
   the proposition [exists M', removal M kxs M'] -- that is, partial sequences
   of removals. And the complete sequences are those where [M'] is empty. This
   is reflected in the definitions of [permitted] and [complete] below.
*)

(* The implementation of [fold] uses two nested loops -- or, more precisely,
   an outer loop over the array, whose body calls the tail-recursive function
   [fold_aux].

   In order to prove that [fold] is correct, we must give a loop invariant for
   each of these two loops. In order to save work, we define a single
   invariant. It is parameterized by [kxs], the previously seen elements of
   the outer loop; by [kys], the previously seen elements of the inner loop;
   and by [i], the outer loop index.

   The invariant expresses the idea that we are VIRTUALLY removing elements
   from the hash table, and building a removal sequence, as we go. It is
   parameterized over [data'] and [M'], which are the physical and logical
   states in which the table would be if the elements were really removed as
   we go. In the loop invariants in [fold], [data'] and [M'] will be
   existentially quantified.

   The elements seen so far by the client are [kxs ++ kys]. So we must have a
   removal sequence [removal M (kxs ++ kys) M']. The client invariant will be
   [I (kxs ++ kys)] -- but it is convenient to keep separate from [iter_inv].

   The invariant is parameterized over [data], which represents the physical
   hash table, and remains unchanged throughout. The data array has been
   virtually changed to [data'], where [data'] is a well-formed data array: it
   satisfies [content] and [no_garbage]. Its length is equal to the length of
   [data]. Furthermore, since we are scanning from left to right and are
   currently at index [i], the array [data'] is empty up to index [i]
   (excluded) and identical to [data] after index [i] (also excluded). At
   index [i], (if this index is valid,) [data'[i]] is the suffix of [data[i]]
   obtained by removing [kys].

   We do NOT own the hash table during iteration (and we do not want to, as
   that would preclude creating multiple iterators on a single table). We
   are happy to just require that it remain in state [data]. As a result,
   the invariant [iter_inv] is pure and can be defined in [Prop].

   This invariant is somewhat complicated. Fortunately, once we have defined
   it, the proof goes through rather easily. Furthermore, the same invariant
   is used in the proof of the iterator (see further on). *)

Section IterInv.

Variable M : model key A.
Variable data : list (MK.bucket_ A).

Definition permitted kxs :=
  exists M', removal M kxs M'.

Definition complete kxs :=
  removal M kxs empty.

Lemma wf_permitted_complete:
  wf (Seqs permitted complete).
Proof.
  unfold wf. simpl. splits.
  (* TEMPORARY the first two could be lemmas about transitive closure *)
  (* [nil] is permitted, because a reflexive-transitive closure is reflexive. *)
  { eexists. eauto with removal. }
  (* [permitted] is a reflexive-transitive closure, hence is prefix-closed. *)
  { intros kxs kzs (?&?) (kys&?). subst kzs.
    forwards (M'&?&?): (removal_split Eequiv). eauto.
    eexists. eauto. }
  (* [complete] implies [permitted]. *)
  { unfold permitted, complete. eauto. }
Qed.

Definition iter_inv data' M' kxs kys i :=
  (* Information that does not change during iteration. *)
  power_of_2 (length data) /\
  content M data /\
  no_garbage data /\
  (* Information that changes during iteration. *)
  0 <= i <= length data /\
  removal M (kxs ++ kys) M' /\
  length data' = length data /\
  content M' data' /\
  no_garbage data' /\
  (forall j, 0 <= j < i -> data'[j] = MK.Void) /\
  (i < length data -> bucket2list data[i] = kys ++ bucket2list data'[i]) /\
  (i = length data -> kys = nil) /\
  (forall j, i < j < length data' -> data'[j] = data[j])
.

Ltac expose_iter_inv :=
  unfold iter_inv at 1;
  intros (
    Hdata & Hcontent & Hnogarbage &
    [ Hi1 Hi2 ] & Hremoval' & Hlen' & Hcontent' & Hnogarbage' &
    Hlo' & Hthere' & Hkys & Hhi'
  ).

(* This lemma is used to establish the invariant. *)

Lemma iter_inv_initialization:
  forall init,
  table_inv M init data ->
  iter_inv data M nil nil 0.
Proof using.
  intros. expose3.
  unfold iter_inv. rew_list. repeat split; eauto with removal.
  { intros. false. math. }
Qed.

(* In the presence of [iter_inv], the conditions [data'[i] = MK.Void] and [kys =
   bucket2list data[i]] are equivalent. They signal the termination of the inner
   loop. *)

Lemma iter_inv_inner_termination:
  forall data' M' kxs kys i,
  iter_inv data' M' kxs kys i ->
  i < length data ->
  (data'[i] = MK.Void) = (kys = bucket2list data[i]).
Proof using.
  do 5 intro. expose_iter_inv. introv Hi.
  specializes Hthere' Hi.
  extens. split; intro H.
  { rewrite H in Hthere'. simpl in Hthere'. rew_list in Hthere'. eauto. }
  { subst kys. eauto using app_cancel_nil_l, bucket2list_injective_nil. }
Qed.

(* These lemmas are used when the inner loop is finished and we are about to
   begin a new iteration of the outer loop. We then move from a state described
   by [kxs, kys, i] to a state described by [kxs ++ kys, nil, i + 1]. *)

Lemma iter_inv_preservation:
  forall data' M' kxs kys i,
  data'[i] = MK.Void ->
  i < length data ->
  iter_inv data' M' kxs kys i ->
  iter_inv data' M' (kxs ++ kys) nil (i + 1) /\
  (i + 1 < length data -> data'[i + 1] = data[i + 1]).
Proof using.
  introv ? ?. expose_iter_inv. split.
  (* Conclusion 1. *)
  { repeat split; rew_list; try solve [ eauto | math ].
    (* The following three proof obligations update the correspondence
       between [data] and [data'] as the index [i] is incremented. Boring. *)
    { intros j [ ? ? ].
      tests: (j < i). { eauto. }
      assert (i = j). math. subst j. eauto. }
    { intros. rewrite Hhi' by math. eauto. }
    { intros j [ ? ? ].
      assert (i < j). math. eauto. }
  }
  (* Conclusion 2. *)
  (* This guarantees that we are ready to read the next bucket. *)
  { intros. rewrite Hhi' by math. eauto. }
Qed.

(* The invariant implies that [kxs ++ kys] is permitted. Furthermore, in
   the special case where iteration is over, the invariant implies that
   [kxs] is complete. *)

Lemma iter_inv_permitted:
  forall data' M' kxs kys i,
  iter_inv data' M' kxs kys i ->
  permitted (kxs ++ kys) /\
  (i = length data -> complete kxs).
Proof using.
  do 5 intro. expose_iter_inv. split.
  (* [permitted] *)
  { unfold permitted. eauto. }
  (* [complete] *)
  { xintros. unfold complete.
    rewrite Hkys in Hremoval' by eauto. rew_list in Hremoval'.
    (* The table is now virtually empty. *)
    assert (M' = empty).
    { eapply func_ext_1. intros k.
      rewrite <- Hcontent'.
      rewrite Hlen'.
      rewrite Hlo' by eauto.
      eauto. }
    (* Thus, we have presented a complete removal sequence to the client. *)
    subst M'. eauto. }
Qed.

(* The invariant implies that [i] is nonnegative. *)

Lemma iter_inv_zero:
  forall data' M' kxs kys i,
  iter_inv data' M' kxs kys i ->
  0 <= i.
Proof using.
  do 5 intro. expose_iter_inv. eauto.
Qed.

(* If the invariant holds and [kys & (k, y)] is a prefix of [data[i]],
   then the pair [k, y] is the first element of [data'[i]]. *)

Lemma iter_inv_extend:
  forall data' M' kxs kys k y i,
  i < length data ->
  prefix (kys & (k, y)) (bucket2list data[i]) ->
  iter_inv data' M' kxs kys i ->
  exists kys', bucket2list data'[i] = (k, y) :: kys'.
Proof using.
  introv Hi Hprefix. expose_iter_inv.
  specializes Hthere' Hi.
  forwards [ kys' ? ]: use_prefix_snoc Hprefix Hthere'.
  eauto.
Qed.

(* If the first element of the virtual bucket [data'[i]] is a pair [k, y]
   then we can present this pair to the client, and the iteration invariant
   is maintained. *)

Lemma iter_inv_next:
  forall data' M' kxs kys i,
  i < length data ->
  forall k y b,
  MK.More k y b = data'[i] ->
  iter_inv data' M' kxs kys i ->
  exists data'', (* data'[i := b] *)
  exists M'',    (* remove M' k   *)
  b = data''[i] /\
  removal M (kxs ++ kys & (k, y)) M'' /\
  iter_inv data'' M'' kxs (kys & (k, y)) i.
Proof using.
  (* Nothing deep in this proof -- just administration. *)
  introv Hi Hcons. expose_iter_inv.
  specializes Hthere' Hi.
  (* Searching for [k] in the bucket [data'[i]] yields [y]. *)
  assert (Some y = head (bfilter k data'[i])).
  { rewrite <- Hcons. simpl. decide_if. reflexivity. }
  (* Therefore, its hash code, modulo, must be [i]. *)
  assert (i = k // data').
  { eapply no_garbage_contrapositive; try rewrite Hlen'; eauto. }
  (* Also, the rest of the bucket can be characterized as follows. *)
  assert (b = bremove k data'[i]).
  { rewrite <- Hcons. simpl. decide_if. reflexivity. }
  (* We may extend the removal sequence with the element [k, y].
     This takes us VIRTUALLY to the logical state [remove M'' k]. *)
  assert (removal M (kxs ++ kys & (k, y)) (remove M' k)).
  { rewrite <- app_assoc.
    eapply removal_append. { eauto. }
    econstructor; [ | eauto | eauto with removal ].
    rewrite <- Hcontent'.
    subst i. eauto. }
  (* [eauto] hints used in the following. *)
  assert (forall j, index data j -> index data' j).
  { intros. eapply index_length_unfold. rewrite Hlen'. eauto. }
  assert (forall j, j < i -> j < length data).
  { subst i. rewrite Hlen'. eauto using lt_trans. }
  assert (forall j, 0 <= i -> i < j -> 0 <= j).
  { intros. math. }
  assert (power_of_2 (length data')).
  { rewrite Hlen'. eauto. }
  (* Provide the witnesses. *)
  exists (data'[i := b]) (remove M' k).
  (* Check that every conjunct in the conclusion is satisfied. *)
  unfold iter_inv.
  rewrite length_update.
  repeat split; try solve [ subst b i; eauto with content no_garbage ];
  intros; unpack;
  repeat rewrite read_update_neq by solve [ eauto | math ];
  repeat rewrite read_update_eq by eauto;
  eauto.
  (* 2 subgoals should be left. *)
  { rew_list. rewrite Hthere'. rewrite <- Hcons. reflexivity. }
  { false. math. }
Qed.

End IterInv.

(* This is to clarify the meaning of [cca_unboxed], which is used in the
   specification of [fold] below. It is just a calling convention that
   says the first and second component of the key-value pair are passed
   separately to the function [f]. *)

Goal
  forall B,
  (@cca_unboxed key A B)
  =
  (fun f kx (accu : B) => app f [(fst kx) (snd kx) accu]).
Proof.
  reflexivity.
Qed.

(* The (public) specification and proof of [fold]. *)

Theorem fold_spec_ro:
  forall M s B I,
  Fold MK.fold
    (@cca_unboxed key A B)
    (permitted M) (complete M) I
    (* [fold] requires and preserves this,
       so does not modify the table: *)
    (fun h => h ~> TableInState M s)
    (* [f] receives this and must preserve
       it, hence can read the table: *)
    (fun h => h ~> TableInState M s).
Proof using.
  intros. intros f h Hf accu.
  xcf. expose2. xstraight.
  match goal with foo: Pervasives_ml.ref_ B |- _ => rename foo into state end. (* TEMPORARY *)
  (* The invariant of the outer loop. [kxs] is existentially quantified,
     while [kys] is [nil], as the inner loop is inactive. [i] is the loop
     index. *)
  xfor_inv (fun i =>
    Hexists kxs data' M' accu,
    h ~> TableInState M s \*
    state ~> Ref accu \*
    I kxs accu \*
    \[ iter_inv M data data' M' kxs nil i ]
  ).
  { math. }

  (* 1. The outer loop invariant holds initially. *)
  { wrapup1. wrapup2. xsimpl*. eauto using iter_inv_initialization. }

  (* 2. The outer loop invariant is preserved. *)
  { clear accu. intros i [ ? ? ].
    xpull. intros kxs data' M' accu Hinv.
    (* Read [accu]. *)
    xapp. xintros.
    (* Read [data[i]]. *)
    xapp_spec* in_state_bucket_read_spec h. xintros.
    (* Call [fold_aux]. *)
    (* The invariant of the inner loop. [kxs] and [i] are now fixed (by the
       outer loop), while [kys] is the loop index, so to speak. *)
    xapp_spec* fold_aux_spec (>>
      (data[i])
      (fun kys accu =>
         Hexists data' M',
         I (kxs ++ kys) accu \*
         h ~> TableInState M s \*
         \[ iter_inv M data data' M' kxs kys i ]
      )
    );
    rew_list in *.

    (* 2.A. Argue that a call to [f] by [fold_aux] has the desired effect
       of extending the removal sequence that we are building. *)
    { clear data' M' accu Hinv.
      intros [ k y ] kys accu Hprefix.
      (* The array [data'] represents the current VIRTUAL state of the hash table. *)
      xpull. intros data' M' Hinv.

      (* The pair [k, y] that is passed to [f] must be the first element of
         the bucket [data'[i]]. *)
      forwards [ kys' Hfirst ]: iter_inv_extend; eauto.
      forwards: nonempty_bucket Hfirst. unpack.

      (* Therefore, we can present this pair to the client, changing [kys] to
         [kys & (k, y)], and our invariant will be preserved. *)
      forwards (data'' & M'' & ? & Hpermitted & ?): iter_inv_next; eauto.

      (* Call [f]. Note that [Hpermitted] guarantees that the precondition of [f]
         is met. *)
      xapply (Hf (k, y) (kxs ++ kys)).
      { unfold permitted. rew_list. eauto. }
      { xsimpl. }

      (* Now reasoning about the state after the call to [f]. *)
      rew_list. xsimpl*.
    }

    (* 2.B. Precondition of [fold_aux]. *)
    { xsimpl*. }

    (* 2.C. Postcondition of [fold_aux]. This is where the inner loop ends
       and the next iteration of the outer loop begins. *)
    { clear data' M' Hinv.
      intros kys _ data' M' Hinv Hkys.
      xapp. (* [state := ...]. *)
      xsimpl (kxs ++ kys).
      forwards Hequiv: iter_inv_inner_termination; eauto.
      rewrite <- Hequiv in Hkys.
      forwards (?&?): iter_inv_preservation; eauto. }
  }

  (* 3. The outer loop invariant finally implies the postcondition. *)
  { clear accu. intros kxs data' M' accu Hinv.
    xapp. (* [!state] *)
    forwards [ Hpermitted Hcomplete ]: iter_inv_permitted. { eauto. }
    rew_list in Hpermitted.
    forwards: Hcomplete. { math. }
    xsimpl* kxs. subst. xsimpl. }
Qed.

(* If the client did not name the current state [s], then the function [f]
   cannot have access (not even read-only access) to the hash table.
   TEMPORARY In fact, we COULD give a spec where the function [f] must be
   polymorphic in [s] and has access to [h ~> TableInState M s]. But it
   would no longer be an instance of [Fold]. Would it be
   worth the trouble? *)

Theorem fold_spec:
  forall M B I,
  Fold MK.fold
    (@cca_unboxed key A B)
    (permitted M) (complete M) I
    (* [fold] requires and preserves this: *)
    (fun h => h ~> Table M)
    (* [f] does not have access to the table: *)
    (fun h => \[]).
Proof using.
  unfold Fold. introv Hf. intros accu.
  xchange* name_current_state; xpull; intros s.
  xapp_spec* fold_spec_ro.
  (* Weaken the spec of [f] by framing out [h ~> TableInState M s]. *)
  { intros. xapply* Hf. }
  xchange* (@forget_current_state c M s).
Qed.

(* -------------------------------------------------------------------------- *)

(* For simplicity, [iter] is defined in terms of [fold]. *)

Lemma iter_spec_ro:
  forall M s I,
  Iter MK.iter
    (@cc_unboxed _ _)
    (permitted M) (complete M)
    I
    (fun h => h ~> TableInState M s)
    (fun h => h ~> TableInState M s).
Proof using.
  unfold Iter, cc_unboxed. introv Hf. xcf.
  xfun. (* define the anonymous function *)
  xapp_spec* fold_spec_ro. unfold cca_unboxed.
  intros [ k x ] kxs accu Hpermitted.
  xapp. (* apply the anonymous function *)
  eapply Hf. eauto.
Qed.

Lemma iter_spec:
  forall M I,
  Iter MK.iter
    (@cc_unboxed _ _)
    (permitted M) (complete M)
    I
    (fun h => h ~> Table M)
    (fun h => \[]).
Proof using.
  unfold Iter, cc_unboxed. introv Hf. xcf.
  xfun. (* define the anonymous function *)
  xapp_spec* fold_spec. unfold cca_unboxed.
  intros [ k x ] kxs accu Hpermitted.
  xapp. (* apply the anonymous function *)
  eapply Hf. eauto.
Qed.

(* -------------------------------------------------------------------------- *)

(* A test of [fold]. An example of client-side reasoning. *)

Lemma population_test_spec:
  forall h M,
  app MK._population_test [h]
    INV  (h ~> Table M)
    POST \[= population M ].
Proof using.
  xcf. xfun. xapp_spec fold_spec (fun (kxs : list (key * A)) accu => \[ accu = length kxs ]).
  { unfold cca_unboxed. intros [ k x ] kxs accu ?.
    xpull. xintros. xstraight. xsimpl. rew_list. math. }
  { rew_list. math. }
  { xpull. intros kxs. xintros.
    unfold complete in *.
    xchange* Table_properties; xpull. intros. unpack. subst.
    xsimpl.
    eapply removal_length_complete; eauto with equiv. }
Qed.

(* -------------------------------------------------------------------------- *)

(* The (public, abstract) predicate [it ~> TableIterator h M kxs] means that
   [it] is a valid iterator on the hash table [h ~> Table M] and that it has
   so far produced the key-value pairs in the list [kxs]. *)

Definition TableIterator h M s kzs (it : MK.iterator_ A) := \[
  exists d data data' M' kxs kys i,
  it = (d, i, data'[i]) /\
  s = (d, data) /\
  i < length data /\
  kxs ++ kys = kzs /\
  iter_inv M data data' M' kxs kys i
].

(* The predicate [it ~> TableIterator h M s kxs] is duplicable. That is, our
   iterators are persistent. *)

Theorem TableIterator_duplicable:
  forall it h M s kxs,
  duplicable (it ~> TableIterator h M s kxs).
Proof using.
  intros. xunfold TableIterator. eauto with duplicable.
Qed.

Hint Resolve TableIterator_duplicable : duplicable.

(* The sequence of elements seen so far is of course permitted. *)

Theorem TableIterator_properties:
  forall it h M s kxs,
  it ~> TableIterator h M s kxs ==> it ~> TableIterator h M s kxs \* \[
    permitted M kxs
  ].
Proof using.
  intros. rename kxs into kzs. xunfold TableIterator.
  xpull. intros (d & data & data' & M' & kxs & kys & i & ? & ? & ? & ? & Hinv).
  xsimpl; [ | do 7 eexists; splits; eauto ].
  forwards [ Hpermitted Hcomplete ]: iter_inv_permitted. { eauto. }
  subst. eauto.
Qed.

(* -------------------------------------------------------------------------- *)

Lemma iterate_spec:
  forall h M s,
  app MK.iterate [h]
    INV  (h ~> TableInState M s)
    POST (fun it => it ~> TableIterator h M s nil).
Proof using.
  xcf. expose. xstraight.
  xunfold TableIterator. wrapup. xsimpl*.
  exists d data data M (@nil (key * A)) (@nil (key * A)). exists 0.
  eauto 15 using iter_inv_initialization with table.
Qed.

(* -------------------------------------------------------------------------- *)

Lemma next_spec_preliminary:
  forall h M s d data data' M',
  s = (d, data) ->
  forall i kxs kys kzs it,
  i < length data ->
  it = (d, i, data'[i]) ->
  kxs ++ kys = kzs ->
  iter_inv M data data' M' kxs kys i ->
  app MK.next [it]
    INV  (h ~> TableInState M s)
    POST (fun o =>
      match o with
      | None =>
          \[ complete M kzs ]
      | Some (k, x, it) =>
          it ~> TableIterator h M s (kzs & (k, x)) \*
          \[ permitted M (kzs & (k, x)) ]
      end).
Proof using.
  (* By well-founded induction on [i], which grows while remaining
     strictly less than [length data]. *)
  introv ?. intro i.
  induction_wf: (int_upto_wf (length data - 1)) i.
  (* Examine the code. *)
  xcf. subst it. xmatch. xval. xmatch.
  (* Case: [More]. *)
  { clear IH. xret.
    (* The pair [k, x] comes first. *)
    assert (Hfirst: bucket2list data'[i] = (k, x) :: bucket2list b).
    { match goal with h: _ = data'[i] |- _ => rewrite <- h end. reflexivity. }
    forwards: nonempty_bucket Hfirst. unpack.
    (* Therefore, we can present this pair to the client. *)
    forwards (data'' & M'' & ? & Hremoval'' & Hinv): iter_inv_next; eauto.
    subst b.
    (* Witnesses. *)
    xunfold TableIterator. subst kzs. xsimpl; rew_list.
    { unfold permitted. eauto. }
    { do 7 eexists. splits; eauto. }
  }
  (* Case: [Nil]. *)
  { (* Increment [i]. *)
    xstraight.
    (* Read [Array.length data]. *)
    xapp_spec* in_state_length_spec. xintros.
    (* Ghost move to the next bucket. This corresponds, logically, to the end
       of the inner loop in [fold]. *)
    forwards [ Hinv Hread ]: iter_inv_preservation data.
    { eauto. } { math. } { eauto. }
    (* Test. *)
    xif.
    (* Sub-case: [i + 1 < length data]. *)
    { (* Read the next bucket. *)
      assert (0 <= i). { eauto using iter_inv_zero. }
      assert (0 <= i + 1). { math. }
      xapp_spec* in_state_bucket_read_spec. xintros.
      (* Call [next] recursively. *)
      subst kzs.
      specializes IH Hinv; rew_list; eauto.
      { unfold upto. math. }
      rewrite Hread in IH by eauto.
      xapp.
    }
    (* Sub-case: [i + 1 = length data]. *)
    { clear IH.
      assert (i + 1 = length data). { math. }
      forwards [ ? Hcomplete ]: iter_inv_permitted; eauto.
      subst kzs. xret. xsimpl*. }
  }
Qed.

Lemma next_spec:
  forall h M s kxs it,
  app MK.next [it]
    INV  (h ~> TableInState M s \* it ~> TableIterator h M s kxs)
    POST (fun o =>
      match o with
      | None =>
          \[ complete M kxs ]
      | Some (k, x, it) =>
          it ~> TableIterator h M s (kxs & (k, x)) \*
          \[ permitted M (kxs & (k, x)) ]
      end).
Proof using.
  intros. rename kxs into kzs.
  xunfold TableIterator at 1. unfold TableIterator at 1.
  xpull. intros (d & data & data' & M' & kxs & kys & i & ? & ? & ? & ? & Hinv).
  xapp_spec next_spec_preliminary; try solve [ eauto | xsimpl ].
  { intros o. xsimpl.
    xunfold TableIterator. xsimpl. do 7 eexists. splits; eauto. }
Qed.

(* -------------------------------------------------------------------------- *)

(* The cascade. *)

(* A cascade is an iterator. However, whereas the iterator (above) is
   described by an ad hoc abstract type ['a iterator] and an ad hoc abstract
   predicate [TableIterator], a cascade is an object of type ['a cascade],
   described by the predicate [Cascade]. Therefore, it offers an interface
   that is the same as that of every other cascade. *)

(* The cascade invariant [cascade_inv h M data i b kzs] contains essentially
   the same information as [TableIterator] above. It uses [iter_inv] to
   express the fact that an iteration is ongoing. It records the fact that the
   elements enumerated so far, [kzs], are in fact [kxs ++ kys], that is, the
   elements enumerated by all previous iterations of the outer loop, followed
   with the elements enumerated by the inner loop in this iteration of the
   outer loop. It records the fact that the current bucket [b] is [data'[i]],
   that is, the remaining suffix of [data[i]]. Finally, it imposes the
   invariant [i < length data], which is not a consequence of [iter_inv], as
   the latter implies [i <= length data]. *)

Definition cascade_inv h M data i b kzs :=
  (* The unused parameter [h] is still useful as a unification hint. *)
  exists data' M' kxs kys,
  b = data'[i] /\
  i < length data /\
  kxs ++ kys = kzs /\
  iter_inv M data data' M' kxs kys i.

Ltac destruct_cascade_inv h :=
  destruct h as (data' & M' & kxs & kys & Hb & Hi & Hseq & Hinv).

(* The internal predicate [cascade_witness h M d data] describes the internal
   state of every possible cascade produced by the auxiliary function
   [cascade_aux]. It is parameterized by [kxs] and [c]. It serves as a witness
   for the existential quantifier [S] in the definition of cascades. *)

Definition cascade_witness h M d data kxs c :=
  (* The address of the data array, [d], does not change as long as the
     cascade is active. On the other hand, the variables [i] and [b]
     change when a new cascade is constructed, so they are existentially
     quantified here. *)
  exists i b,
  cascade_inv h M data i b kxs /\
  (* The following two lines state that the function call [c()] has the same
     effect as the function call [cascade_aux d i b]. Thus, they express the
     fact that the closure [c] is [fun () -> cascade_aux d i b]. *)
  forall P (Q : Cascade_ml.head_ (key * A) -> hprop),
  app MK.cascade_aux [d i b] P Q -> app c [tt] P Q.

Ltac intros_cascade_witness :=
  intros (i & b & Hinv & Hspec).

(* The specification of the auxiliary function [cascade_aux] does not yet
   involve the predicate [Cascade]. Instead, it is expressed in terms of
   [cascade_inv] and [cascade_witness]. *)

Lemma cascade_aux_spec:
  forall h M (d : array (MK.bucket_ A)) data i b kxs,
  cascade_inv h M data i b kxs ->
  app MK.cascade_aux [d i b]
    INV  (h ~> TableInState M (d, data))
    POST (fun o => \[
      match o with
      | Cascade_ml.Nil =>
          complete M kxs
      | Cascade_ml.Cons kx c =>
          cascade_witness h M d data (kxs & kx) c
      end]).
Proof using.
  (* By well-founded induction on [i], which grows while remaining strictly
     less than [length data]. This induction is made necessary by the fact
     that [cascade_aux] is recursive; when it finds that the bucket [b] is
     empty, it immediately calls itself recursively. *)
  intros h M d data i.
  induction_wf: (int_upto_wf (length data - 1)) i.
  intros b kzs Hinv. destruct_cascade_inv Hinv.
  (* Examine the code. *)
  xcf. xval; subst. xmatch.
  (* Case: [More]. *)
  { clear IH. xfun. xret.
    (* The pair [k, x] comes first. *)
    assert (Hfirst: bucket2list data'[i] = (k, x) :: bucket2list b).
    { match goal with h: _ = data'[i] |- _ => rewrite <- h end. reflexivity. }
    forwards: nonempty_bucket Hfirst. unpack.
    (* Therefore, we can present this pair to the client. *)
    forwards (data'' & M'' & ? & ? & ?): iter_inv_next; eauto.
    subst b.
    (* Witnesses. *)
    xsimpl. rew_list. exists i (data''[i]). split.
    { econstructor. do 3 eexists; splits; eauto. }
    (* The Hoare triple that describes the next cascade. *)
    { introv Hself. xapp. eauto. }
  }
  (* Case: [Nil]. *)
  { (* Increment [i]. *)
    xret. xintros.
    (* Read [Array.length data]. *)
    xapp_spec* in_state_length_spec h. xintros.
    (* Ghost move to the next bucket. This corresponds, logically, to the end
       of the inner loop in [fold]. *)
    forwards [ ? Hread ]: iter_inv_preservation data; eauto.
    (* Test. *)
    xif.
    (* Sub-case: [i + 1 < length data]. *)
    { (* Read the next bucket. *)
      assert (0 <= i). { eauto using iter_inv_zero. }
      assert (0 <= i + 1). { math. }
      xapp_spec* in_state_bucket_read_spec. xintros.
      (* Call [cascade_aux] recursively. *)
      xapp.
      { unfold upto. math. }
      { econstructor.
        rewrite <- Hread by eauto.
        exists M' (kxs ++ kys) (@nil (key * A)). rew_list. eauto. }
    }
    (* Sub-case: [i + 1 = length data]. *)
    { clear IH.
      assert (i + 1 = length data). { math. }
      forwards [ ? Hcomplete ]: iter_inv_permitted; eauto.
      xret. xsimpl*. }
  }
Qed.

(* We can now prove that the function [cascade] produces a cascade, which
   remains valid as long as the hash table remains in state [s]. *)

Theorem cascade_spec:
  forall h M s,
  app MK.cascade [h]
    INV  (h ~> TableInState M s)
    POST (fun c =>
      c ~> Cascade
        (h ~> TableInState M s)
        (Seqs (permitted M) (complete M))
        nil
    ).
Proof using.
  (* First, examine the code of [cascade]. *)
  intros h M s.
  xcf. expose. subst s. xstraight. xfun. xret.
  xunfold TableInState at 2. unfold TableInState at 2.
  xsimpl*; eauto 8 with table.
  (* There remains to check that the final call to [cascade_aux] produces
     a valid cascade. The lemma [cascade_aux_spec] (above) provides the
     bulk of the proof; an application of the lemma [pure_cascade] helps
     package it as a cascade. *)
  pure_cascade.
  exists (cascade_witness h M d data); splits.
  (* [permitted] *)
  { intros kzs c. intros_cascade_witness. destruct_cascade_inv Hinv.
    forwards [ ? _ ]: iter_inv_permitted; eauto. subst kzs. eauto. }
  (* The invariant holds initially. *)
  { exists 0 (data[0]). split.
    { econstructor. exists M (@nil (key * A)) (@nil (key * A)).
      splits; eauto 8 using iter_inv_initialization with table. }
    { intros. xapp. eauto. }
  }
  (* The invariant is preserved. *)
  { intros kxs c. intros_cascade_witness.
    (* A call to [c] reduces to a call to [cascade_aux], *)
    eapply Hspec.
    (* which we know preserves the cascade invariant. *)
    xapply* cascade_aux_spec. }
Qed.

(* -------------------------------------------------------------------------- *)

(* Preliminary work on [memo]. *)

(* TEMPORARY [memo] should be placed in a separate module. 1. It could work
   with any kind of map that supports create/add/find with the same spec.
   2. The proof should involve monotonicity, so we can guarantee that two
   calls to the memoized function with the same argument produce the same
   result, even if the relation [Q x y] is not functional. *)

(* TEMPORARY [AffineFun] should move to CFML *)

Definition AffineFun A B (P : A -> hprop) (Q : A -> B -> hprop) f :=
  Hexists S, S \* \[ forall x,
    app f [x]
      PRE  (         S \* P x   )
      POST (fun y => S \* Q x y )
  ].

Lemma use_affinefun:
  forall A B (P : A -> hprop) (Q : A -> B -> hprop) f x,
  app f [x]
    PRE  (         f ~> AffineFun P Q \* P x   )
    POST (fun y => f ~> AffineFun P Q \* Q x y ).
Proof using.
  intros. xunfold AffineFun. xpull. intros S Hf. xapp. xsimpl*.
Qed.

Lemma add_here:
  forall M k v,
  (add M k v) k = v :: M k.
Proof (* using Eequiv *).
  intros. unfold add. decide_if. eauto.
Qed.

(* TEMPORARY unfinished *)
Lemma memo_spec:
  forall f,
  forall P : key -> hprop,
  forall (Q : key -> A -> hprop),
  (forall x1 x2 y, E x1 x2 -> Q x1 y ==> Q x2 y) ->
  (forall x y, duplicable (Q x y)) ->
  (forall x,
     app f [x]
       PRE  (         P x  )
       POST (fun y => Q x y)
  ) ->
  app MK.memo [f]
    \[]
    (fun f' => f' ~> AffineFun P Q).
Local Hint Resolve use_None_head.
Proof using.
  introv Hproper Hdup Hf. xcf. xapp. xfun. xret.
  xunfold AffineFun.
  (* The invariant is as follows. At any time, [h] is a hash table of domain
     [D] and of content [M], where [M] is lean, and we own the separating
     conjunction of the [Q x y], where [x] ranges over [D] and [y] is the
     image of [x] through [M]. *)
  set (I := Hexists D M,
    h ~> Table M \*
    \[ is_domain D M ] \*
    \[ lean M ] \*
    fold sep_monoid (fun x => Hexists y, Q x y \* \[ head (M x) = Some y ]) D
  ).
  xsimpl I.
  (* The invariant holds initially. *)
  { unfold I. xsimpl (\{} : set key) empty.
    { rewrite fold_empty. simpl. xsimpl. }
    { eauto with lean. }
    { eauto with equiv is_domain. }
  }
  (* The invariant is preserved. *)
  intros. unfold I at 1. xpull. intros D M Hdomain Hlean. xapp.
  (* Code for the anonymous function. *)
  xapp_spec find_spec. xintros. xmatch.
  (* Case: [None]. *)
  { xret. unfold I.
    (* TEMPORARY should duplicate [Q x y] for the particular [x] we are interested in *)
    xsimpl*. skip. }
  (* Case: [Some]. *)
  { assert (~ represented E x D). { skip. } (* TEMPORARY *)
    assert (D \# \{x}). { skip. } (* TEMPORARY a consequence *)
    xapp_spec Hf.
    xapp_spec add_spec. xintros.
    xdup (Q x y).
    xret.
    unfold I.
    xsimpl* (D \u \{x}) (add M x y).
    rewrite fold_union; eauto 2 with finite typeclass_instances.
    2: eauto with is_domain equiv. (* TEMPORARY ugly *)
    simpl. rewrite fold_single.
    xsimpl* y.
    { erewrite fold_pointwise. xsimpl. xsimpl. eauto with typeclass_instances.
      skip. skip. eauto with finite.
      intros x' Hx'. simpl.
      (* TEMPORARY messy. rewrite add_elsewhere. *)
      skip. }
    { rewrite add_here. eauto. }
    { eauto with typeclass_instances. }
  }
Qed.

End S.

End MakeSpec.
