Set Implicit Arguments.
Require Import Coq.Setoids.Setoid.    (* [rewrite] *)
Require Import Coq.Classes.Morphisms. (* [Proper] *)
Require Import LibTactics.
Require Import LibLogic.
Require Import LibRelation.
Require Import LibSet.

(* -------------------------------------------------------------------------- *)

(* Facilities for working with an equivalence relation. *)

Section F.

Variables (A : Type) (E : binary A) (Eequiv : equiv E).

Hint Resolve (equiv_refl Eequiv) (equiv_sym Eequiv) (equiv_trans Eequiv) : equiv.

Lemma Erefl: forall k, E k k.
Proof. eauto with equiv. Qed.

Lemma Esym: forall k1 k2, E k1 k2 -> E k2 k1.
Proof. eauto with equiv. Qed.

Lemma Etrans: forall k1 k2 k3, E k1 k2 -> E k2 k3 -> E k1 k3.
Proof. eauto with equiv. Qed.

Global Instance equiv_Equivalence : Equivalence E.
Proof. econstructor; eauto with equiv. Qed.

End F.

(* Erefl, Esym, Etrans are not accepted as hints. But they can be
   explicitly applied by the client to an equivalence relation
   and (then) declared as hints. *)

(* -------------------------------------------------------------------------- *)

(* A function [M : A -> B] is compatible with two equivalence relations [EA]
   and [EB] iff it maps equivalent elements to equivalent results. This can
   be stated as: [Proper (EA ==> EB) M]. *)

(* -------------------------------------------------------------------------- *)

(* A little theory of sets up to [E]. *)

Section S.

Variables (A : Type) (E : binary A) (Eequiv : equiv E).

Hint Resolve (equiv_refl Eequiv) (equiv_sym Eequiv) (equiv_trans Eequiv) : equiv.

Implicit Type S : set A.

(* [represented x S] holds if [x] is in [S], up to [E]. *)

Definition represented x S :=
  exists y, E x y /\ y \in S.

Hint Unfold represented : equiv.

Lemma member_represented:
  forall x S,
  x \in S ->
  represented x S.
Proof.
  intros. exists x. eauto with equiv.
Qed.

Lemma member_represented_contrapositive:
  forall x S,
  ~ represented x S ->
  x \notin S.
Proof.
  repeat intro; eauto using member_represented.
Qed.

Lemma represented_contradiction:
  forall x y S,
  ~ represented x S ->
  represented y S ->
  E x y ->
  False.
Proof.
  unfold represented. introv Hr [ z [ ? ? ]] ?.
  rew_logic in Hr.
  specializes Hr z.
  rew_logic in Hr.
  branches; eauto with equiv.
Qed.

Hint Resolve represented_contradiction : equiv.

Lemma not_represented_union_inv_1:
  forall x S1 S2,
  ~ represented x (S1 \u S2) ->
  ~ represented x S1.
Proof.
  unfold represented. introv Hnr. rew_logic in *. intros y.
  specializes Hnr y. set_norm. tauto.
Qed.

Lemma not_represented_union_inv_2:
  forall x S1 S2,
  ~ represented x (S1 \u S2) ->
  ~ represented x S2.
Proof.
  unfold represented. introv Hnr. rew_logic in *. intros y.
  specializes Hnr y. set_norm. tauto.
Qed.

Lemma not_represented_singleton_inv_lax:
  forall x y,
  ~ represented x \{y} ->
  ~ E x y.
Proof.
  unfold represented. introv Hnr. rew_logic in *.
  specializes Hnr y. set_norm. tauto.
Qed.

Lemma not_represented_singleton_inv_strict:
  forall x,
  ~ represented x \{x} ->
  False.
Proof.
  eauto using not_represented_singleton_inv_lax with equiv.
Qed.

Lemma not_represented_union:
  forall x S1 S2,
  ~ represented x S1 ->
  ~ represented x S2 ->
  ~ represented x (S1 \u S2).
Proof using Eequiv.
  unfold represented. introv H1 H2. rew_logic in *.
  intros y. specializes H1 y. specializes H2 y. rew_logic in *.
  destruct H1; destruct H2; eauto; right.
  set_norm. rew_logic. tauto.
Qed.

Lemma not_represented_diff_inv:
  forall j k k' S,
  ~ represented j (S \-- k') ->
  ~ E j k ->
  E k k' ->
   ~ represented j S.
Proof using Eequiv.
  unfold represented. introv H1 H2 H3. rew_logic in *.
  intros y. specializes H1 y. rew_logic in *.
  destruct H1 as [ | H1 ]; eauto.
  set_norm. rew_logic in *.
  destruct H1 as [ H1 | ]; eauto. subst y. left.
  eauto with equiv.
Qed. (* phew *)

Lemma represented_diff_inv_1:
  forall x S1 S2,
  represented x (S1 \- S2) ->
  represented x S1.
Proof.
  introv [ y [ ? ? ]]. set_norm. exists y. tauto.
Qed.

(* [irredundant S] holds if [S] holds at most one element of each
   equivalence class of [E]. *)

Definition irredundant S :=
  forall x y, x \in S -> y \in S -> E x y -> x = y.

Lemma represented_diff_inv_2:
  forall x S1 S2,
  represented x (S1 \- S2) ->
  S2 \c S1 ->
  irredundant S1 ->
  ~ represented x S2.
Proof.
  introv [ y [ ? ? ]] Hincl Hirr [ z [ ? ? ]]. set_norm; unpack.
  forwards: Hirr y z. { eauto. } { eauto. } { eauto with equiv. }
  subst. tauto.
Qed.

Lemma not_represented_diff:
  forall x S1 S2,
  irredundant S1 ->
  represented x S2 ->
  S2 \c S1 ->
  ~ represented x (S1 \- S2).
Proof.
  (* This is a contrapositive of the previous lemma. *)
  repeat intro. eapply represented_diff_inv_2; eauto.
Qed.

Lemma not_represented_diff_singleton:
  forall x y S,
  irredundant S ->
  E x y ->
  y \in S ->
   ~ represented x (S \- \{y}).
Proof.
  eauto 8 using not_represented_diff, single_incl_r with equiv typeclass_instances.
Qed.

Lemma not_represented_diff_singleton_corollary:
  forall x y S,
  irredundant S ->
  E x y ->
  y \in S ->
  forall z,
  z \in S \- \{y} ->
  ~ E x z.
Proof.
  intros. forwards: not_represented_diff_singleton; eauto.
  intro. eapply represented_contradiction; eauto with equiv.
Qed.

Lemma irredundant_empty:
  irredundant \{}.
Proof.
  repeat intro. set_norm. false.
Qed.

Lemma irredundant_extend:
  forall x S,
  irredundant S ->
  ~ represented x S ->
  irredundant (S \u \{x}).
Proof.
  introv Hirr Hnr. intros y z. repeat intro. set_norm.
  (* 4 branches, 2 of which are immediate, 2 of which are impossible. *)
  repeat branches; subst; eauto; false; eauto 6 with equiv.
Qed.

Lemma irredundant_included:
  forall S1 S2,
  irredundant S2 ->
  S1 \c S2 ->
  irredundant S1.
Proof.
  unfold irredundant. introv Hirr Hincl Hx Hy HE.
  forwards: Hirr HE.
  { set_norm. eauto. }
  { set_norm. eauto. }
  eauto.
Qed.

Lemma irredundant_diff:
  forall S1 S2,
  irredundant S1 ->
  irredundant (S1 \- S2).
Proof.
  intros.
  assert (S1 \- S2 \c S1). { set_norm. intros. set_norm. tauto. }
  eauto using irredundant_included.
Qed.

End S.

(* Specialized versions of [represented_contradiction], for use by
   [eauto]. The hypothesis [equiv E] must not come first: as [E]
   does not appear in the conclusion, [eauto] has trouble guessing
   it. *)

Lemma represented_contradiction_1:
  forall A (E : binary A) x y S,
  ~ represented E x S ->
  equiv E ->
  represented E y S ->
  E x y ->
  False.
Proof. eauto using represented_contradiction. Qed.

Lemma represented_contradiction_2:
  forall A (E : binary A) x y S,
  represented E y S ->
  equiv E ->
  ~ represented E x S ->
  E x y ->
  False.
Proof. eauto using represented_contradiction. Qed.

(* The above two lemmas can be used as [eauto] hints, but this
   is costly. We restrict their use to the case where their
   first hypothesis already exists in the context. *)

Hint Extern 1 (False) =>
  match goal with
  | h: ~ represented _ _ _ |- _ =>
      eapply (represented_contradiction_1 h)
  | h: represented _ _ _ |- _ =>
      eapply (represented_contradiction_2 h)
  end : equiv.

Hint Unfold represented : equiv.

Hint Resolve
  member_represented
  not_represented_union_inv_1 not_represented_union_inv_2
  not_represented_singleton_inv_lax
  member_represented_contrapositive
: equiv.

Hint Resolve
  irredundant_empty
  irredundant_extend
  irredundant_diff
: irredundant.
