Set Implicit Arguments.
Require Import LibTactics LibLogic.
Require Import LibInt.

(* A definition of ``being a power of two'' in [Z]. *)

Definition power_of_2 (n : int) :=
  exists k, 0 <= k /\ n = 2 ^ k.

(* A power of two is positive. *)

Lemma power_of_2_positive:
  forall n,
  power_of_2 n ->
  0 < n.
Proof.
  intros n [ k [ ? ? ]]. subst.
  int_comp_to_zarith. (* @arthur: your overloading of the math operators is painful! *)
  apply Z.pow_pos_nonneg; omega.
Qed.

(* Doubling a power of two yields a power of two. *)

Lemma power_of_2_double:
  forall x,
  power_of_2 x ->
  power_of_2 (x * 2).
Proof.
  unfold power_of_2.
  intros x [ k [ ? ? ]]. exists (k + 1). subst. split.
  { math. }
  { rewrite pow2_succ by eauto. math. }
Qed.

(* Doubling a power of two yields a larger power of two. *)

Lemma power_of_2_double_larger:
  forall x y,
  power_of_2 x ->
  x * 2 <= y ->
  x <= y.
Proof.
  intros x y Hx Hxy.
  forwards: power_of_2_positive Hx. clear Hx.
  math.
Qed.

(* The sequence of the powers of two goes to infinity. *)

Lemma powers_of_2_diverge:
  forall x n,
  0 < x ->
  exists k,
  0 <= k /\ n <= x * 2 ^ k.
Proof.
  intros x n Hx.
  tests: (0 < n).
  (* Case: [0 < n]. *)
  { exists (Z.succ (Z.log2 n)). (* not tight... *)
    split.
    { generalize (@Z.log2_nonneg n). math. }
    { forwards: (@Z.log2_spec n). { math. }
      rewrite <- (@Z.mul_1_l n).
      apply Zmult_le_compat.
      { math. }
      { rewrite Z.mul_1_l. math. }
      { math. }
      { math. }
  }}
  (* Case: [n <= 0]. *)
  { exists 0. split; simpl; math. }
Qed.

(* -------------------------------------------------------------------------- *)

(* Theory of logical [and]. *)

(* TEMPORARY prove these axioms *)

(* If [y] is a power of two, then for every [x], the bitwise AND of [x] and
   [y - 1] is comprised between 0 inclusive and [y] exclusive. *)

(* We do not use the modulo operator at all. Reasoning about modulo is not
   convenient when possibly-negative numbers are involved. *)

Axiom land_in_bounds_left:
  forall x y,
  power_of_2 y ->
  0 <= Z.land x (y - 1).

Axiom land_in_bounds_right:
  forall x y,
  power_of_2 y ->
  Z.land x (y - 1) < y.
