
(** Vectors (aka resizable arrays, growing arrays, dynamic arrays, etc.)

    This module implements arrays that automatically expand as necessary.
    Its implementation uses a traditional array and replaces it with a
    larger array when needed (and elements are copied from the old array
    to the new one). The current implementation doubles the capacity when
    growing the array (and shrinks it whenever the number of elements
    comes to one fourth of the capacity).

    The unused part of the internal array is filled with a dummy value,
    which is user-provided at creation time (and referred to below
    as ``the dummy value''). Consequently, vectors do not retain pointers
    to values that are not used anymore after a shrinking.

    Vectors provide an efficient implementation of stacks, with a
    better locality of reference than list-based implementations (such
    as standard library {!Stack}). A stack interface is provided,
    similar to that of {!Stack} (though {!Vector.push} have arguments
    in the other way round). Inserting [n] elements with
    {!Vector.push} has overall complexity O(n) i.e. each insertion has
    amortized constant time complexity. *)

type 'a t
  (** The polymorphic type of vectors.
      This is a mutable data type. *)


(** {2 Operations proper to vectors, or with a different type and/or
    semantics than those of module [Array]} *)

val create: ?capacity:int -> dummy:'a -> 'a t
(** [Vector.make c dummy] returns a fresh vector of length [0].
   All the elements of this new vector are initially
   physically equal to [dummy] (in the sense of the [==] predicate).

   Raise [Invalid_argument] if [c < 0] or [c > Sys.max_array_length].
   If the value of [dummy] is a floating-point number, then the maximum
   size is only [Sys.max_array_length / 2].

   When [capacity] is omitted, it defaults to 0. *)

val make: ?dummy:'a -> int -> 'a -> 'a t

val init: dummy:'a -> int -> (int -> 'a) -> 'a t
(** [Vector.init n f] returns a fresh vector of length [n],
   with element number [i] initialized to the result of [f i].
   In other terms, [Vector.init n f] tabulates the results of [f]
   applied to the integers [0] to [n-1].

   Raise [Invalid_argument] if [n < 0] or [n > Sys.max_array_length].
   If the return type of [f] is [float], then the maximum
   size is only [Sys.max_array_length / 2].*)

val resize: 'a t -> int -> unit
(** [Vector.resize a n] sets the length of vector [a] to [n].

   The elements that are no longer part of the vector, if any, are
   internally replaced by the dummy value of vector [a], so that they
   can be garbage collected when possible.

   Raise [Invalid_argument] if [n < 0] or [n > Sys.max_array_length]. *)

(** {2 Sequence interface} *)

include SequenceSig.ClearLength with type 'a t := 'a t

(** {2 Stack interface}

    Contrary to standard library's {Stack}, module {Vector} uses less space
    (between N and 2N words, instead of 3N) and has better data locality. *)

include StackSig.PushPopTop with type 'a t := 'a t

(** {2 Array interface} *)

include SequenceSig.GetSet with type 'a t := 'a t

include SequenceSig.SubFillBlit with type 'a t := 'a t

include SequenceSig.IterMap with type 'a t := 'a t
include SequenceSig.IteriMapi with type 'a t := 'a t
include SequenceSig.FoldLeftRight with type 'a t := 'a t

val append: 'a t -> 'a t -> 'a t
(** [Vector.append a1 a2] returns a fresh vector containing
    the concatenation of the elements of [a1] and [a2].

    It works correctly even if [a1] and [a2] are the same vector. *)

val merge_right: 'a t -> 'a t -> unit
  (** [Vector.merge_right a1 a2] moves all elements of [a2] to the end
      of [a1]. Empties [a2]. Assumes [a1] and [a2] to be disjoint.

      FIXME: say something about the capacity. *)

(* FIXME: remove to_list, of_list, to_array, of_array ? *)

val to_list : 'a t -> 'a list
(** [Vector.to_list a] returns the list of all the elements of [a]. *)

val of_list: dummy:'a -> 'a list -> 'a t
(** [Vector.of_list dummy l] returns a fresh vector containing the elements
    of [l]. *)

val to_array: 'a t -> 'a array
(** [Vector.to_array a] returns the array of all the elements of [a]. *)

val of_array: dummy:'a -> 'a array -> 'a t
(** [Vector.of_array dummy a] returns a fresh vector containing the elements
    of [a]. *)

(** {2 Only if you know what you are doing...} *)
(* FIXME
val unsafe_get : 'a t -> int -> 'a
val unsafe_set : 'a t -> int -> 'a -> unit
*)

