

(*-----------------------------------------------------------------------------*)

(** TODO: update

    Representation of a bootstrapped stack:
    -- front-outer buffer: where top elements are pushed 
    -- front-inner buffer: elements after front-outer
    -- middle: an option on a bootstrapped stack of chunks, 
       storing deeper elements in the stack.

    The front-inner buffer is always either empty or full.
    The structure is empty iff the front-outer buffer is empty.
    If middle is not None, then it must store at least one chunk. *)

type 'a t = { 
  mutable size : int;
  mutable tail_data : 'a list;
  mutable outer_data : 'a array;
  mutable aux : ('a aux) option;
  chunk_capacity_pow : int; 
  tail_capacity_pow : int;
  default : 'a; }

   (* TODO: two capacity fields could be merged using arithmetic *)

and 'a aux = {
  mutable inner_full : bool;
  mutable inner_data : 'a array;
  mutable middle : ('a array) t; }



(*-----------------------------------------------------------------------------*)

(* takes a number, expected to be a power of two, and returns the 
   corresponding exponent *)

let rec pow_of_int n =
   assert (n > 0);
   if n = 1 then 0 else begin
      assert (n mod 2 = 0);
      1 + (pow_of_int (n/2))
   end

(** Creation *)

let create chunk_capacity tail_capacity d = 
   assert (tail_capacity >= chunk_capacity);
   { size = 0;
     tail_data = [];
     outer_data = [||]; 
     aux = None;
     chunk_capacity_pow = pow_of_int chunk_capacity;
     tail_capacity_pow = pow_of_int tail_capacity;
     default = d; }

(** Emptiness test *)    
  
let is_empty s = 
  (s.size = 0)

(** Push front *)

let rec push : 'a. 'a -> 'a t -> unit = fun x s ->
   let tail_capacity = (1 lsl s.tail_capacity_pow) in
   if s.size < tail_capacity then begin
      s.tail_data <- x::s.tail_data;
      s.size <- s.size + 1;
      if s.size = tail_capacity then begin
         s.outer_data <- Array.make (1 lsl s.chunk_capacity_pow) s.default; 
      end
   end else begin
      let i = (s.size land ((1 lsl s.chunk_capacity_pow) - 1)) in
      if i = 0 && s.size > tail_capacity then shift_outer s;
      s.outer_data.(i) <- x;
      s.size <- s.size + 1;
  end

and shift_outer : 'a. 'a t -> unit = fun s ->
  begin match s.aux with
  | None ->
      let default_rec = Array.make 0 s.default in
      s.aux <- Some { 
         inner_full = true;
         inner_data = s.outer_data;
         middle = create (1 lsl s.chunk_capacity_pow) (1 lsl s.tail_capacity_pow) default_rec; };
      s.outer_data <- Array.make (1 lsl s.chunk_capacity_pow) s.default;
  | Some aux ->
      let fi = aux.inner_full in
      let di = aux.inner_data in
      aux.inner_data <- s.outer_data;
      aux.inner_full <- true;
      if fi then begin
         s.outer_data <- Array.make (1 lsl s.chunk_capacity_pow) s.default;
         push di aux.middle
      end else begin
         if Array.length di = 0 then begin
            s.outer_data <- Array.make (1 lsl s.chunk_capacity_pow) s.default;
         end else begin
            s.outer_data <- di;
         end
      end
   end

(** Pop front *)

let rec pop : 'a. 'a t -> 'a = fun s -> 
   let tail_capacity = (1 lsl s.tail_capacity_pow) in
   if s.size <= tail_capacity then begin
      s.size <- s.size - 1;
      if s.size = (tail_capacity / 2) && s.aux <> None then begin
         s.outer_data <- [||];
         s.aux <- None;
      end;
      match s.tail_data with
      | x::q -> s.tail_data <- q; x
      | [] -> assert false
   end else begin
      let i = ((s.size-1) land ((1 lsl s.chunk_capacity_pow) - 1)) in
      let x = s.outer_data.(i) in
      s.outer_data.(i) <- s.default;
      s.size <- s.size - 1;
      if i = 0 && s.size > tail_capacity then unshift_outer s;
      x
  end

and unshift_outer : 'a. 'a t -> unit = fun s ->
  begin match s.aux with
  | None -> ()
  | Some aux ->
      if aux.inner_full then begin
         let d = s.outer_data in
         s.outer_data <- aux.inner_data;
         aux.inner_data <- d;
         aux.inner_full <- false;
      end else begin
         s.outer_data <- pop aux.middle;
         if is_empty aux.middle then begin
            aux.inner_data <- [||];
         end;
      end
   end

(** Top

let top : 'a. 'a t -> 'a = fun s -> 
   let co = s.front_outer in
   assert (not (Chunk.is_empty co));
   let x = Chunk.pop co in
   x
 *)

(*-----------------------------------------------------------------------------*)


(*-----------------------------------------------------------------------------*)


let print_chunk f c =
   Printf.printf "[";
   for k = 0 to pred (Array.length c) do
      f c.(k);
   done;
   Printf.printf "] "

let rec print : 'a. ('a -> unit) -> 'a t -> unit = fun f s ->
   Printf.printf "{\n ch=2^%d; tl=2^%d; size= %d; fo= " s.chunk_capacity_pow s.tail_capacity_pow s.size;
   print_chunk f s.outer_data;
   Printf.printf ";\n ";
   begin match s.aux with
     | None -> Printf.printf "mi=None"
     | Some aux -> 
        Printf.printf ";\n ff=%d; fi= " (if aux.inner_full then 1 else 0);
        print_chunk f aux.inner_data;
        print (fun c -> print_chunk f c) aux.middle
   end;
   Printf.printf ";\n tail= ";
   List.iter f s.tail_data;
   Printf.printf "}\n ";
