
let unsome = function
   | None -> assert false
   | Some x -> x

(*-----------------------------------------------------------------------------*)

module Chunk = FixedCapaStack

(** Aliases *)

type 'a chunk = 'a Chunk.t 


(*-----------------------------------------------------------------------------*)

(** Representation of a bootstrapped stack:
    -- front-outer buffer: where top elements are pushed 
    -- front-inner buffer: elements after front-outer
    -- middle: an option on a bootstrapped stack of chunks, 
       storing deeper elements in the stack
    -- tail: a list of length nbtail, which stores the items
       at the rear of the sequence.

    The front-inner buffer is always either empty or full.
    The structure is empty iff the tail is empty.
    If middle is not None, then it must store at least one chunk. 
    If front-outer stores any item, then the tail must have length nbtail_max. *)

and 'a t = { 
  mutable front_outer : 'a chunk;
  mutable front_inner : 'a chunk;
  mutable middle : (('a chunk) t) option;
  mutable nbtail : int;
  mutable tail : 'a list; 
  nbtail_max : int }


(*-----------------------------------------------------------------------------*)

(** Creation *)

let create capacity tail_capacity d = 
  assert (tail_capacity > 0);
  let co = Chunk.create capacity d in
  let ci = Chunk.create capacity d in
  { front_outer = co;
    front_inner = ci;
    middle = None;
    nbtail = 0;
    tail = [];
    nbtail_max = tail_capacity }

(** Default *)    

let default s =
   Chunk.default s.front_outer

(** Capacity *)    

let capacity s =
   Chunk.capacity s.front_outer

(** Emptiness test *)    
  
let is_empty s = 
  s.nbtail = 0

(** Push front *)

let rec push : 'a. 'a -> 'a t -> unit = fun x s ->
   if s.nbtail < s.nbtail_max then begin
      assert (Chunk.is_empty s.front_outer 
           && Chunk.is_empty s.front_inner 
           && s.middle = None);
      s.nbtail <- s.nbtail + 1;
      s.tail <- x :: s.tail;
   end else begin
      if Chunk.is_full s.front_outer then begin
         let ci = s.front_inner in
         assert (Chunk.is_full ci || Chunk.is_empty ci);
         s.front_inner <- s.front_outer;
         if Chunk.is_empty ci then begin
            s.front_outer <- ci;
         end else begin
            s.front_outer <- Chunk.create (capacity s) (default s);
            if s.middle = None then begin
               let default_rec = Chunk.create 0 (default s) in
               let tail_capacity_rec = capacity s in (* FREE TO CHOOSE *)
               let m = create (capacity s) tail_capacity_rec default_rec in
               s.middle <- Some m;
            end;
            push ci (unsome s.middle);
         end
      end;
      Chunk.push x s.front_outer
   end

(** Pop front *)

let rec pop : 'a. 'a t -> 'a = fun s -> 
   let co = s.front_outer in
   if not (Chunk.is_empty co) then begin
      let x = Chunk.pop co in
      if Chunk.is_empty co then begin
         let ci = s.front_inner in
         assert (Chunk.is_full ci || Chunk.is_empty ci);
         if Chunk.is_full ci then begin
            s.front_outer <- ci;
            s.front_inner <- co;
         end else begin
            match s.middle with
            | None -> () 
            | Some m ->
               s.front_outer <- pop m;
               if is_empty m then begin
                  s.middle <- None;
               end
         end;
      end;
      x
   end else begin
      assert (Chunk.is_empty s.front_outer 
           && Chunk.is_empty s.front_inner 
           && s.middle = None);
      match s.tail with
      | x::q -> 
         s.tail <- q;
         s.nbtail <- s.nbtail - 1;
         x
      | [] -> assert false
   end

(** Top *)

let top : 'a. 'a t -> 'a = fun s -> 
   let co = s.front_outer in
   if not (Chunk.is_empty co) then begin
      let x = Chunk.pop co in
      x
   end else begin
      match s.tail with
      | x::q -> x
      | [] -> assert false
  end


(*-----------------------------------------------------------------------------*)

(* INEFFICIENT
let rec to_list : 'a. 'a t -> 'a list = fun s ->
     (Chunk.to_list s.front_outer)
   @ (Chunk.to_list s.front_inner)
   @ (match s.middle with
     | None -> []
     | Some m -> (List.concat (List.map (fun c -> Chunk.to_list c) (to_list m))))
   @ s.tail
 *) 

let print_chunk f c =
   Printf.printf "[";
   List.iter f (Chunk.to_list c);
   Printf.printf "] "

let rec print : 'a. ('a -> unit) -> 'a t -> unit = fun f s ->
   Printf.printf "{\n fo= ";
   print_chunk f s.front_outer;
   Printf.printf ";\n fi= ";
   print_chunk f s.front_inner;
   Printf.printf ";\n mi= ";
   begin match s.middle with
     | None -> Printf.printf "None"
     | Some m -> print (fun c -> print_chunk f c) m
   end;
   Printf.printf ";\n nbtail=%d " s.nbtail;
   Printf.printf ";\n tail= ";
   List.iter f s.tail;
   Printf.printf "}\n ";

