
let unsome = function
   | None -> assert false
   | Some x -> x

(*-----------------------------------------------------------------------------*)

module Chunk = HardFixedCapaStack

(** Aliases *)

type 'a chunk = 'a Chunk.t 


(*-----------------------------------------------------------------------------*)

(** Representation of a bootstrapped stack:
    -- front-outer buffer: where top elements are pushed 
    -- front-inner buffer: elements after front-outer
    -- middle: an option on a bootstrapped stack of chunks, 
       storing deeper elements in the stack.

    The front-inner buffer is always either empty or full.
    The structure is empty iff the front-outer buffer is empty.
    If middle is not None, then it must store at least one chunk. *)

and 'a t = { 
  mutable front_outer : 'a chunk;
  mutable front_inner : 'a chunk;
  mutable middle : (('a chunk) t) option; }


(*-----------------------------------------------------------------------------*)

(** Creation *)

let create capacity d = 
  let co = Chunk.create capacity d in
  let ci = Chunk.create capacity d in
  { front_outer = co;
    front_inner = ci;
    middle = None }

(** Default *)    

let default s =
   Chunk.default s.front_outer

(** Capacity *)    

let capacity s =
   Chunk.capacity s.front_outer

(** Emptiness test *)    
  
let is_empty s = 
  Chunk.is_empty s.front_outer

(** Push front *)

let rec push : 'a. 'a -> 'a t -> unit = fun x s ->
   if Chunk.is_full s.front_outer then begin
      let ci = s.front_inner in
      assert (Chunk.is_full ci || Chunk.is_empty ci);
      s.front_inner <- s.front_outer;
      if Chunk.is_empty ci then begin
         s.front_outer <- ci;
      end else begin
         s.front_outer <- Chunk.create (capacity s) (default s);
         if s.middle = None then begin
            s.middle <- Some (create (capacity s) (Chunk.create 0 (default s)));
         end;
         push ci (unsome s.middle);
      end
   end;
   Chunk.push x s.front_outer

(** Pop front *)

let rec pop : 'a. 'a t -> 'a = fun s -> 
   let co = s.front_outer in
   assert (not (Chunk.is_empty co));
   let x = Chunk.pop co in
   if Chunk.is_empty co then begin
      let ci = s.front_inner in
      assert (Chunk.is_full ci || Chunk.is_empty ci);
      if Chunk.is_full ci then begin
         s.front_outer <- ci;
         s.front_inner <- co;
      end else begin
         match s.middle with
         | None -> ()  (* structure is now empty *)
         | Some m ->
            s.front_outer <- pop m;
            if is_empty m then begin
               s.middle <- None;
            end
      end;
   end;
   x

(** Top *)

let top : 'a. 'a t -> 'a = fun s -> 
   let co = s.front_outer in
   assert (not (Chunk.is_empty co));
   let x = Chunk.pop co in
   x


(*-----------------------------------------------------------------------------*)

