Set Implicit Arguments.
Require Import Coq.Setoids.Setoid.    (* [rewrite] *)
Require Import Coq.Classes.Morphisms. (* [Proper] *)
Require Import Coq.Program.Equality.  (* [dependent destruction] *)
Require Import LibCore.
Require Import LibSet.
Require Import LibStruct.
Require Import LibEquiv.
Require Import LibListZ.
Require Import LibEpsilon.

(* TEMPORARY clean up / move these tactics *)

Ltac decide_if :=
  (case_if; try solve [ false; eauto with equiv ]); [ idtac ].

Lemma refl_eq: forall A, refl (@eq A).
Proof. repeat intro; eauto. Qed.

Ltac rewrite_under_fold H :=
  (rewrite fold_pointwise; eauto using refl_eq with typeclass_instances finite); [ |
    intros; erewrite H by eauto; reflexivity ].

Ltac isolate k :=
  rewrite fold_isolate with (x := k) by eauto with typeclass_instances finite.

Ltac de_isolate k :=
  match goal with |- context[?add ?lhs (fold ?monoid ?f ?K)] =>
    change add with (monoid_oper monoid) at 1; (* not quite sure about this occurrence number, of course *)
    change lhs with (f k);
    rewrite <- fold_isolate with (x := k) by eauto with typeclass_instances finite
  end.

(* -------------------------------------------------------------------------- *)

(* Z as a monoid. *) (* TEMPORARY move to TLC *)

Notation Zmonoid := (monoid_ Zplus 0).

Global Instance Zmonoid_monoid:
  Monoid Zmonoid.
Proof using.
  constructor; repeat intro; math.
Qed.

Global Instance Zmonoid_commutative:
  Monoid_commutative Zmonoid.
Proof using.
  constructor. typeclass. repeat intro; math.
Qed.

Notation Zsum K f := (fold Zmonoid f K).

(* -------------------------------------------------------------------------- *)

Section S.

Variables (K : Type) (E : binary K) (Eequiv : equiv E) (V : Type).

Hint Resolve Eequiv : equiv.
Hint Resolve (Erefl Eequiv) (Esym Eequiv) (Etrans Eequiv) : equiv.

Notation represented := (represented E).
Notation irredundant := (irredundant E).

(* -------------------------------------------------------------------------- *)

(* The logical model. *)

Definition model := K -> list V.

Implicit Type k : K.
Implicit Type v : V.
Implicit Type D : set K.
Implicit Type M : model.

Definition empty : model :=
  fun _ => nil.

Definition add M k v : model :=
  fun k' =>
    If E k' k then v :: M k' else M k'.

Definition remove M k : model :=
  fun k' =>
    If E k' k then tail (M k') else M k'.

Definition nil_outside D M :=
  forall k, ~ represented k D -> M k = nil.

Definition is_domain D M :=
  finite D /\
  nil_outside D M /\
  irredundant D.

Definition is_population M pop :=
  exists D,
  is_domain D M /\
  pop = Zsum D (fun k => length (M k)).

Definition population M :=
  epsilon (is_population M).

(* A hash table is lean if it maps every key to at most one value, i.e.,
   the possibility of binding a single key several times is not exploited. *)

Definition lean M :=
  forall k, length (M k) <= 1.

(* -------------------------------------------------------------------------- *)

(* Lemmas about models. *)

(* Either [M] is [nil] everywhere, or [M] maps some key [k] to a nonempty list
   of values. *)

Lemma model_empty_or_nonempty:
  forall M,
  M = empty \/ exists k, M k <> nil.
Proof using.
  intros.
  tests: (exists k, M k <> nil); [ right; eauto | left ].
  eapply func_ext_1; intros k.
  specializes C k.
  rewrite not_not in C.
  eauto.
Qed.

(* The empty model is empty everywhere. *)

Lemma head_empty:
  forall k x,
  head (empty k) = Some x ->
  False.
Proof using.
  simpl. congruence.
Qed.

(* The empty model is lean. *)

Lemma lean_empty:
  lean empty.
Proof.
  intros k. unfold empty. unfold length, LibList.length. simpl. math.
Qed.

(* When querying [add k x] at [k'], either [k] and [k'] are equal, in
   which case one finds [x], or they are distinct, in which case this
   is equivalent to querying [M] at [k']. *)

Lemma head_add:
  forall M k k' x x', head (add M k x k') = Some x' ->
  E k k' /\ x = x' \/
  head (M k') = Some x'.
Proof using Eequiv.
  unfold add. intros. case_if; simpl in *; [ left | right ].
  { split. eauto with equiv. congruence. }
  { eauto. }
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [add]. *)

(* Adding at a key [k] does not affect an unrelated key [k']. *)

Lemma add_elsewhere:
  forall M k v k',
  ~ E k k' ->
  (add M k v) k' = M k'.
Proof using Eequiv.
  intros. unfold add. decide_if. eauto.
Qed.

(* Adding at an existing key [k] does not affect an unrelated key [j]. *)

Lemma add_existing_elsewhere:
  forall D M v k k',
  irredundant D ->
  E k k' ->
  k' \in D ->
  forall j, j \in D \-- k' ->
  add M k v j = M j.
Proof using Eequiv.
  intros. eapply add_elsewhere. intro.
  set_norm. intuition eauto with equiv.
Qed.

(* Adding at a new key [k] does not affect an existing key [j]. *)

Lemma add_new_elsewhere:
  forall D M k v,
  ~ represented k D ->
  forall j, j \in D ->
  add M k v j = M j.
Proof using Eequiv.
  intros. unfold add. decide_if. reflexivity.
Qed.

(* Adding at a new key preserves leanness. *)

Lemma lean_add:
  forall M k v,
  lean M ->
  Proper (E ==> eq) M ->
  M k = nil ->
  lean (add M k v).
Proof using Eequiv.
  introv Hlean Hproper Hk. intro k'. unfold add. case_if.
  (* Case: [k'] is equivalent to [k]. *)
  { rewrite Hproper, Hk by eauto. unfold length, LibList.length. simpl. math. }
  (* Case: [k'] is not equivalent to [k]. *)
  { eauto. }
Qed.

(* [add] is compatible with [E]. *)

Lemma proper_add:
  forall M k v,
  Proper (E ==> eq) M ->
  Proper (E ==> eq) (add M k v).
Proof using Eequiv.
  unfold add. intros. intros k1 k2 HE.
  case_if; decide_if; rewrite HE; eauto.
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [remove]. *)

(* Removing at [k] does not affect an unrelated key [j]. *)

Lemma remove_elsewhere:
  forall D M k j,
  ~ E k j ->
  remove M k j = M j.
Proof using Eequiv.
  unfold remove. intros. decide_if. reflexivity.
Qed.

(* If [remove M k] and [M] agree at key [k], then they agree at every key. *)

Lemma remove_unchanged:
  forall M k,
  remove M k k = M k ->
  Proper (E ==> eq) M ->
  remove M k = M.
Proof.
  unfold remove. intros. decide_if.
  eapply func_ext_1. intros k'. case_if as HE; [ | reflexivity ].
  assert (HMk: M k = nil). { eauto using tail_self_inv. }
  rewrite HE, HMk. reflexivity.
Qed.

(* If [remove M k] and [M] do not agree at key [k], then [M k] is non-nil. *)

Lemma remove_changed:
  forall M k,
  remove M k k <> M k ->
  M k <> nil.
Proof.
  unfold remove. intros. decide_if. introv h. rewrite h in *. tauto.
Qed.

Lemma remove_changed_reciprocal:
  forall M k,
  M k <> nil ->
  remove M k k <> M k.
Proof.
  unfold remove. intros. decide_if. introv h.
  forwards: tail_self_inv h.
  tauto.
Qed.

(* [remove] is compatible with [E]. *)

Lemma proper_remove:
  forall M k,
  Proper (E ==> eq) M ->
  Proper (E ==> eq) (remove M k).
Proof using Eequiv.
  unfold remove. intros. intros k1 k2 HE.
  case_if; decide_if; rewrite HE; eauto.
Qed.

(* [remove] preserves leanness. *)

Lemma lean_remove:
  forall M k,
  lean M ->
  lean (remove M k).
Proof using Eequiv.
  introv Hlean. intro k'.
  forwards: Hlean k'.
  unfold remove. case_if.
  { rewrite length_tail_le. math. }
  { eauto. }
Qed.

(* If [M] is lean, then [remove M k] maps [k] to [nil]. *)

Lemma lean_remove_find:
  forall M k,
  lean M ->
  remove M k k = nil.
Proof using Eequiv.
  unfold remove. intros. decide_if. eauto using length_le_1_implies_nil_tail.
Qed.

(* Removing followed with adding (also known as [replace]) preserves leanness. *)

Lemma lean_replace:
  forall M k v,
  lean M ->
  Proper (E ==> eq) M ->
  lean (add (remove M k) k v).
Proof using Eequiv.
  eauto using lean_add, lean_remove, proper_remove, lean_remove_find.
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [nil_outside]. *)

Lemma nil_outside_empty:
  nil_outside \{} empty.
Proof using.
  repeat intro. reflexivity.
Qed.

Lemma nil_outside_add_existing_key:
  forall D M k v,
  nil_outside D M ->
  represented k D ->
  nil_outside D (add M k v).
Proof using Eequiv.
  intros. intros j Hj. unfold add. decide_if. eauto.
Qed.

Lemma nil_outside_add_new_key:
  forall D M k v,
  nil_outside D M ->
  nil_outside (D \u \{k}) (add M k v).
Proof using Eequiv.
  intros. intros j Hj. unfold add. decide_if. eauto 3 with equiv.
Qed.

Lemma nil_outside_remove:
  forall D M k,
  nil_outside D M ->
  nil_outside D (remove M k).
Proof using Eequiv.
  unfold nil_outside, remove. introv Hnil. intros j Hj.
  rewrite Hnil by eauto. case_if; reflexivity.
Qed.

Lemma nil_outside_remove_lean:
  forall D M k k',
  nil_outside D M ->
  lean M ->
  E k k' ->
  nil_outside (D \- \{k'}) (remove M k).
Proof using Eequiv.
  unfold nil_outside, remove, lean. introv Hnil Hlean HE. intros j Hj.
  case_if.
  (* Case: [j] is equivalent to [k] and [k']. Use leanness. *)
  { eauto using length_le_1_implies_nil_tail. }
  (* Case: [j] is not equivalent to [k] and [k']. *)
  { rewrite Hnil by eauto using not_represented_diff_inv. eauto. }
Qed.

Lemma nil_outside_inv:
  forall D M k,
  nil_outside D M ->
  M k <> nil ->
  represented k D.
Proof.
  unfold nil_outside. intros.
  eapply assume_not; intros; false; eauto.
Qed.

Hint Resolve
  nil_outside_empty
  nil_outside_add_existing_key
  nil_outside_add_new_key
  nil_outside_remove
  nil_outside_remove_lean
: nil_outside.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [is_domain]. *)

Lemma is_domain_finite:
  forall D M,
  is_domain D M ->
  finite D.
Proof using.
  unfold is_domain. tauto.
Qed.

Local Hint Resolve is_domain_finite.

Lemma is_domain_empty:
  is_domain \{} empty.
Proof using Eequiv.
  splits; eauto with finite nil_outside irredundant.
Qed.

Lemma is_domain_add_existing:
  forall D M k v,
  is_domain D M ->
  represented k D ->
  is_domain D (add M k v).
Proof using Eequiv.
  unfold is_domain. intuition eauto with nil_outside.
Qed.

Lemma is_domain_add_new:
  forall D M k v,
  is_domain D M ->
  ~ represented k D ->
  is_domain (D \u \{k}) (add M k v).
Proof using Eequiv.
  unfold is_domain. intuition eauto with finite nil_outside irredundant.
Qed.

Lemma is_domain_add:
  forall D M k v,
  is_domain D M ->
  exists D, is_domain D (add M k v).
Proof using Eequiv.
  intros.
  tests: (represented k D).
  { eauto using is_domain_add_existing. }
  { eauto using is_domain_add_new. }
Qed.

Lemma is_domain_remove:
  forall D M k,
  is_domain D M ->
  is_domain D (remove M k).
Proof using Eequiv.
  unfold is_domain. intuition eauto with finite nil_outside irredundant.
Qed.

Lemma is_domain_remove_lean:
  forall D M k k',
  is_domain D M ->
  lean M ->
  E k k' ->
  is_domain (D \-- k') (remove M k).
Proof using Eequiv.
  unfold is_domain. intuition eauto with finite nil_outside irredundant.
Qed.

Hint Resolve
  is_domain_empty
  is_domain_add_existing
  is_domain_add_new
  is_domain_remove
: is_domain.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [is_population]. *)

Lemma is_population_empty:
  is_population empty 0.
Proof using Eequiv.
  exists \{}. split.
  { eauto with is_domain. }
  { rewrite fold_empty. reflexivity. }
Qed.

Lemma is_population_add:
  forall M k v pop,
  is_population M pop ->
  is_population (add M k v) (pop + 1).
Proof using Eequiv.
  introv (D & Hdomain & Hsum).
  destruct (classic (represented k D)) as [ C | C ].
  (* Case: [k] is covered by [D]. That is, the key [k], or some equivalent key,
     is already in the domain [D]. Then, the domain remains [D]. *)
  { exists D. split; [ eauto with is_domain | ].
    (* The population has increased by one. Indeed, the length of [M k'] has
       increased by one, and the length of [M j] elsewhere is unchanged. *)
    destruct Hdomain as (?&?&?).
    destruct C as (k'&?&?).
    isolate k'. simpl.
    unfold add at 1. decide_if.
    rewrite length_cons, <- Zplus_assoc, Zplus_comm. f_equal.
    rewrite_under_fold add_existing_elsewhere.
    de_isolate k'.
    assumption.
  }
  (* Case: [k] is not covered by [D]. That is, neither the key [k] nor any
     equivalent key is in the domain [D]. Then, the domain should now be
     the union of [D] and the singleton set [\{k}]. *)
  { exists (D \u \{k}). split; [ eauto with is_domain | ].
    (* The population has increased by one. Indeed, the length of [M k] has
       increased by one, and the length of [M j] elsewhere is unchanged. *)
    destruct Hdomain as (?&Hcomplete&?).
    isolate k. simpl.
    unfold add at 1. decide_if.
    rewrite length_cons, <- Zplus_assoc, Zplus_comm. f_equal.
    rewrite <- set_add_remove by eauto with equiv.
    rewrite_under_fold add_new_elsewhere.
    rewrite Hcomplete, length_nil, Zplus_0_l by eauto.
    assumption.
  }
Qed.

Local Hint Resolve (not_represented_diff_singleton_corollary Eequiv).
  (* used by [rewrite_under_fold] below. *)

Lemma is_population_remove:
  forall M k pop pop',
  Proper (E ==> eq) M ->
  is_population M pop ->
  pop' = (If remove M k k = M k then pop else pop - 1) ->
  is_population (remove M k) pop'.
Proof using Eequiv.
  introv Hproper Hpopulation ?.
  case_if; subst pop'.
  (* Case: the population has not changed. *)
  { erewrite remove_unchanged by eauto. eauto. }
  (* Case: the population has decreased. *)
  { destruct Hpopulation as (D & Hdomain & Hsum).
    exists D. split; [ eauto with is_domain | ].
    destruct Hdomain as (?&Hcomplete&?).
    (* [k] must be represented in [D], that is, related to some key [k'] in [D]. *)
    assert (HMk: M k <> nil). { eauto using remove_changed. }
    assert (Hr: represented k D). { eauto using nil_outside_inv. }
    destruct Hr as (k'&HE&?).
    isolate k'. simpl.
    unfold remove at 1. decide_if.
    rewrite HE in HMk.
    rewrite length_tail by eauto.
    match goal with |- _ = (?l - 1) + ?r => replace ((l - 1) + r) with ((l + r) - 1) by math end.
    f_equal.
    rewrite_under_fold remove_elsewhere.
    de_isolate k'.
    assumption.
  }
Qed.

(* A corollary of the above lemma. *)

Lemma is_population_remove_strict:
  forall M k pop,
  Proper (E ==> eq) M ->
  is_population M pop ->
  M k <> nil ->
  is_population (remove M k) (pop - 1).
Proof using Eequiv.
  intros.
  forwards: remove_changed_reciprocal; eauto.
  eapply is_population_remove; eauto.
  decide_if. eauto.
Qed.

(* The population of [M] must be nonnegative. *)

Lemma is_population_nonnegative:
  forall M pop,
  is_population M pop ->
  0 <= pop.
Proof using Eequiv.
  introv (D&?&?). subst. eapply fold_induction; eauto with typeclass_instances; simpl.
  { math. }
  { intros s k Hs. math. } (* TEMPORARY why does this work? does [math] know that [length (M k)] is nonnegative? *)
Qed.

(* If [M] is empty, then its population must be zero. *)

Lemma population_uniqueness_empty:
  forall pop,
  is_population empty pop ->
  pop = 0.
Proof using Eequiv.
  unfold empty. introv (D&?&?). subst.
  eapply fold_induction; eauto with typeclass_instances; simpl; intros; rew_list; math.
Qed.

(* The population of [M] is uniquely defined, regardless of which representative
   is chosen in each equivalence class. *)

Lemma population_uniqueness:
  forall pop1 M pop2,
  is_population M pop1 ->
  is_population M pop2 ->
  Proper (E ==> eq) M ->
  pop1 = pop2.
Proof using Eequiv.
  (* By well-founded induction on [pop1]. *)
  intro. induction_wf: (int_downto_wf 0) pop1.
  introv Hpopulation1 Hpopulation2 Hproper.
  (* By cases: *)
  destruct (model_empty_or_nonempty M) as [ Hempty | (k&?) ].
  (* Case: [M] is empty. *)
  { rewrite Hempty in Hpopulation1, Hpopulation2.
    forwards: population_uniqueness_empty Hpopulation1.
    forwards: population_uniqueness_empty Hpopulation2.
    math. }
  (* Case: [M] maps some key [k] to a nonempty list. *)
  { forwards h1: is_population_remove_strict Hpopulation1; eauto.
    forwards h2: is_population_remove_strict Hpopulation2; eauto.
    forwards: is_population_nonnegative h1; eauto.
    forwards: IH h1 h2.
    { unfold downto. math. }
    { eauto using proper_remove. }
    math. }
Qed.

(* If [M] is lean, then its population is the cardinal of a minimal domain
   of [M]. The following lemma prove one half of this statement. *)

Lemma lean_population:
  forall pop D M,
  is_domain D M ->
  Proper (E ==> eq) M ->
  lean M ->
  is_population M pop ->
  pop <= Z.of_nat (card D).
Proof using Eequiv.
  intro. induction_wf: (int_downto_wf 0) pop. introv Hdomain Hproper Hlean Hpop.
  destruct (model_empty_or_nonempty M) as [ | [ k ? ]].
  (* Case: [M] is empty. Then, [pop] is zero. *)
  { subst. forwards: population_uniqueness_empty; eauto. math. }
  (* Case: [M] is nonempty. Apply the induction hypothesis to [remove M k].
     Because [M] is lean, [D \ k'] is a domain of [remove M k], where [k']
     is equivalent to [k] and lies in [D]. *)
  { forwards: is_population_remove_strict; eauto.
    forwards: is_population_nonnegative; eauto.
    forwards (k'&?&?): nil_outside_inv. { eapply Hdomain. } { eauto. }
    forwards fact: IH (pop - 1) (D \-- k') (remove M k);
      eauto using proper_remove, lean_remove, is_domain_remove_lean.
    { unfold downto. math. }
    rewrite card_diff_single in fact by eauto.
    forwards: card_nonempty D; eauto.
    rewrite Nat2Z.inj_sub in fact by eauto.
    math. }
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas about [population]. *)

(* The value [population M] satisfies [is_population M] only if [M] has a finite
   domain [D], and is uniquely defined only if [Proper (E ==> eq) M] holds. Thus,
   this value really is ``the population'' of [M] only when both conditions are
   met. *)

Lemma population_exists:
  forall D M,
  is_domain D M ->
  is_population M (population M).
Proof using.
  intros. unfold population.
  eapply epsilon_spec.
  econstructor. eauto.
Qed.

Lemma population_empty:
  population empty = 0.
Proof using Eequiv.
  unfold population.
  eapply epsilon_elim; eauto using is_population_empty, population_uniqueness_empty.
Qed.

Lemma population_add:
  forall D M k v,
  is_domain D M ->
  Proper (E ==> eq) M ->
  population (add M k v) = population M + 1.
Proof using Eequiv.
  unfold population at 1. intros.
  eapply epsilon_elim.
  { eauto using is_population_add, population_exists. }
  { intros pop Hpop.
    eapply population_uniqueness; eauto using proper_add, is_population_add, population_exists. }
Qed.

Lemma population_remove:
  forall D M k,
  is_domain D M ->
  Proper (E ==> eq) M ->
  population (remove M k) = (If remove M k k = M k then population M else population M - 1).
Proof using Eequiv.
  unfold population at 1. intros.
  eapply epsilon_elim.
  { eauto using is_population_remove, population_exists. }
  { intros pop Hpop.
    eapply population_uniqueness; eauto using proper_remove, is_population_remove, population_exists. }
Qed.

Lemma population_remove_strict:
  forall D M k v,
  is_domain D M ->
  Proper (E ==> eq) M ->
  M k <> nil ->
  population (remove M k) = population M - 1.
Proof using Eequiv.
  intros.
  erewrite population_remove by eauto.
  forwards: remove_changed_reciprocal; eauto.
  decide_if. eauto.
Qed.

Lemma population_nonnegative:
  forall D M,
  is_domain D M ->
  0 <= population M.
Proof using Eequiv.
  intros. unfold population.
  eapply epsilon_elim; eauto using population_exists, is_population_nonnegative.
Qed.

(* -------------------------------------------------------------------------- *)

(* The proposition [removal M kxs M'], where [kxs] is a list of key-value
   pairs, means that, starting from [M], one can successively remove each
   of the key-value pairs in the list [kxs], to obtain [M']. *)

(* This is the labelled reflexive-transitive closure of a labelled relation.
   This notion could be defined in a library. TEMPORARY *)

Inductive removal : model -> list (K * V) -> model -> Prop :=
| RemovalNil:
    forall M,
    removal M nil M
| RemovalCons:
    forall M'' M' M k x kxs,
    Some x = head (M k) ->
    M' = remove M k ->
    removal M' kxs M'' ->
    removal M ((k, x) :: kxs) M''.

Hint Constructors removal : removal.

(* Two removal sequences can be concatenated. *)

Lemma removal_append:
  forall M kxs M' kys M'',
  removal M kxs M' ->
  removal M' kys M'' ->
  removal M (kxs ++ kys) M''.
Proof.
  induction 1; simpl; intros; rew_list; eauto with removal.
Qed.

(* A removal sequence can be split. *)

Lemma removal_split:
  forall kxs kys M M'',
  removal M (kxs ++ kys) M'' ->
  exists M',
  removal M kxs M' /\
  removal M' kys M''.
Proof.
  induction kxs; simpl; introv h; rew_list in *.
  { eauto with removal. }
  { dependent destruction h.
    forwards (M'&?&?): IHkxs. eauto.
    eauto with removal. }
Qed.

(* Domain is preserved by removal. (The most precise domain shrinks.) *)

Lemma removal_domain:
  forall D M,
  is_domain D M ->
  forall kxs M',
  removal M kxs M' ->
  is_domain D M'.
Proof using Eequiv.
  induction 2; subst; eauto with is_domain.
Qed.

Hint Resolve removal_domain : is_domain.

(* The length of a removal sequence is the difference between the
   populations of [M] and [M']. *)

Lemma removal_length_difference:
  forall D M,
  is_domain D M ->
  Proper (E ==> eq) M ->
  forall kxs M',
  removal M kxs M' ->
  population M = length kxs + population M'.
Proof using Eequiv.
  induction 3; rew_list.
  { math. }
  { subst M'.
    forwards: IHremoval; eauto using proper_remove with is_domain.
    forwards: population_remove_strict D M k x; eauto using use_Some_head.
    math. }
Qed.

(* The length of a removal sequence out of [M] is at most the population
   of [M]. *)

Lemma removal_length_at_most:
  forall D M,
  is_domain D M ->
  Proper (E ==> eq) M ->
  forall kxs M',
  removal M kxs M' ->
  length kxs <= population M.
Proof using Eequiv.
  intros.
  forwards: removal_length_difference; eauto.
  forwards: population_nonnegative M'; eauto with is_domain.
  math.
Qed.

(* The length of a complete removal sequence out of [M] is the population
   of [M]. *)

Lemma removal_length_complete:
  forall D M,
  is_domain D M ->
  Proper (E ==> eq) M ->
  forall kxs,
  removal M kxs empty ->
  length kxs = population M.
Proof using Eequiv.
  intros.
  forwards h: removal_length_difference; eauto.
  rewrite population_empty in h.
  math.
Qed.

(* TEMPORARY also, prove that a removal sequence contains no duplicate
   keys (up to [E]) if the table is lean *)

(* -------------------------------------------------------------------------- *)

End S.

Hint Resolve
  is_domain_empty
  is_domain_add
  is_domain_add_new
  is_domain_remove
  removal_domain
: is_domain.

Hint Resolve
  population_empty
  population_remove
: population.

Hint Resolve
  lean_empty
: lean.

Hint Constructors removal : removal.

Hint Resolve is_domain_finite : finite.
